import argparse
import logging
import pathlib
import threading
import time
from collections.abc import Collection
from concurrent.futures import Future, ThreadPoolExecutor
from ipaddress import IPv4Address
from typing import Any, Dict, Final, List, Set, Tuple, Union, cast

import yaml

from quel_ic_config import Quel1Box, Quel1BoxType, QuelClockMasterV1

logger = logging.getLogger()

NUM_CONNECT_RETRY: Final[int] = 2
NUM_LINKUP_RETRY: Final[int] = 3
DEFAULT_MAX_ALLOWABLE_COUNTERDELTA: Final[int] = 256  # = 2048ns


def validate_collection_of_index(v: Collection[int]):
    is_valid: bool = True
    if isinstance(v, Collection):
        for i in v:
            if not isinstance(i, int) or i < 0:
                is_valid = False
                break
    else:
        is_valid = False
    return is_valid


def validate_options_of_box(box_opt: dict[str, Any]) -> bool:
    is_valid: bool = True
    for k, v in box_opt.items():
        if k in {"ignore_crc_error_of_mxfe", "ignore_access_failure_of_adrf6780", "ignore_lock_failure_of_lmx2594"}:
            if not validate_collection_of_index(v):
                logger.error(f"invalid list of indices: '{v}'")
                is_valid = False
        else:
            logger.error(f"invalid option '{k}'")
            is_valid = False
    return is_valid


def validate_boxes_conf(box_confs) -> bool:
    valid_conf: bool = True
    box_names: Set[str] = set()  # Notes: for checking duplicated box_names
    ipaddrs: Set[IPv4Address] = set()  # Notes: for checking duplicated addresses

    # Notes: validating box_confs
    if not isinstance(box_confs, list):
        logger.error("unexpected format of box descriptions")
        return False

    for idx, spec in enumerate(box_confs):
        if not isinstance(spec, dict):
            logger.error(f"unexpected format of descriptions of the {idx}-th box")
            valid_conf = False
            continue

        if "name" not in spec:
            logger.error(f"no 'name' found at the {idx}-th box")
            valid_conf = False
            continue

        box_name = spec["name"]
        if box_name in box_names:
            logger.error(f"duplicated name '{box_name}' at the {idx}-th box")
            valid_conf = False
            continue
        box_names.add(box_name)

        if "ipaddr" not in spec:
            logger.error(f"no 'ipaddr' found of a box '{box_name}'")
            valid_conf = False
            continue

        try:
            box_ipaddr = IPv4Address(spec["ipaddr"])
        except Exception:
            logger.error(f"invalid value '{spec['ipaddr']}' for 'ipaddr' of a box '{box_name}'")
            valid_conf = False
            continue

        if box_ipaddr in ipaddrs:
            logger.error(f"IP address {box_ipaddr:s} of a box '{box_name}' is duplicated")
            valid_conf = False
            continue

        ipaddrs.add(box_ipaddr)

        if "boxtype" not in spec:
            logger.error(f"no 'boxtype' found of a box '{box_name}'")
            valid_conf = False
            continue

        try:
            _ = Quel1BoxType.fromstr(spec["boxtype"])
        except KeyError:
            logger.error(f"invalid boxtype '{spec['boxtype']}' of a box '{box_name}'")
            valid_conf = False
            continue

        if "options" in spec and not validate_options_of_box(spec["options"]):
            logger.error(f"invalid option '{spec['options']}' of a box '{box_name}'")
            valid_conf = False
            continue

    return valid_conf


def validate_clockmaster_conf(cm_conf) -> bool:
    if not (isinstance(cm_conf, list) and len(cm_conf) == 1 and isinstance(cm_conf[0], dict)):
        logger.error("unexpected format of clockmaster descriptions")
        return False

    if "ipaddr" not in cm_conf[0]:
        logger.error("no 'ipaddr' found of clock master")
        return False

    return True


# Notes: no leakage of exceptions is allowed. this function is submitted to pool.
#        all the possible exceptions are handled within function for better diagnosis.
def load_box(box_conf: dict[str, Any]) -> tuple[str, Union[Quel1Box, str]]:
    box_name = box_conf["name"]
    threading.current_thread().name = box_name  # Notes: thread name is printed in log messages.

    box_ipaddr = IPv4Address(box_conf["ipaddr"])
    box_type = Quel1BoxType.fromstr(box_conf["boxtype"])
    box_opt: dict[str, Any] = box_conf.get("options", {})

    exception: str = ""
    for _ in range(NUM_CONNECT_RETRY):
        try:
            retval: Union[Quel1Box, str] = Quel1Box.create(
                ipaddr_wss=str(box_ipaddr),
                ipaddr_sss=str(box_ipaddr + 0x010000),
                ipaddr_css=str(box_ipaddr + 0x040000),
                boxtype=box_type,
                ignore_crc_error_of_mxfe=box_opt.get("ignore_crc_error_of_mxfe", set()),
                ignore_access_failure_of_adrf6780=box_opt.get("ignore_access_failure_of_adrf6780", set()),
                ignore_lock_failure_of_lmx2594=box_opt.get("ignore_lock_failure_of_lmx2594", set()),
            )
            logger.info(f"connected to {box_name} successfully")
            break
        except Exception as e:
            exception = str(e.args[0])
            logger.warning(f"failed to connect to {box_name} due to exception: '{e}'")
    else:
        retval = exception
        logger.error(f"give up to connect to {box_name} due to repeated failures")

    return box_name, retval


def load_conf_v1(conf, filename: pathlib.Path) -> tuple[bool, list[dict[str, str]]]:
    if len(set(conf.keys()) - {"version", "boxes"}) != 0:
        logger.error(f"CANCELED: unexpected keys are found in the config file '{filename}'")

    if "boxes" not in conf:
        logger.error(f"CANCELED: no 'boxes' key is found in the config file '{filename}'")
        return False, []

    if not validate_boxes_conf(conf["boxes"]):
        logger.error(f"CANCELED: broken configuration file '{filename}', quitting")
        return False, []

    return True, conf["boxes"]


def load_conf_v2(conf, filename: pathlib.Path) -> tuple[bool, list[dict[str, str]], dict[str, str]]:
    if len(set(conf.keys()) - {"version", "clockmaster", "boxes"}) != 0:
        logger.error(f"CANCELED: unexpected keys are found in the config file '{filename}'")

    if "clockmaster" not in conf:
        logger.error(f"CANCELED: no 'clockmaster' key is found in the config file '{filename}'")
        return False, [], {}

    if "boxes" not in conf:
        logger.error(f"CANCELED: no 'boxes' key is found in the config file '{filename}'")
        return False, [], {}

    if not (validate_boxes_conf(conf["boxes"]) and validate_clockmaster_conf(conf["clockmaster"])):
        logger.error(f"CANCELED: broken configuration file '{filename}', quitting")
        return False, [], {}

    return True, conf["boxes"], conf["clockmaster"][0]


def load_conf(filename: pathlib.Path) -> Tuple[bool, int, list[dict[str, str]], dict[str, str]]:
    try:
        with open(filename) as f:
            conf = yaml.safe_load(f)
    except Exception as e:
        logger.error(f"CANCELED: failed to load '{filename}' due to exception: '{e}'")
        return False, 0, [], {}

    if conf is None:
        logger.error(f"CANCELED: empty config file '{filename}'")
        return False, 0, [], {}

    if "version" not in conf:
        logger.error(f"CANCELED: version is not specified in the config file '{filename}'")
        return False, 0, [], {}

    version: int = conf["version"]
    if version == 1:
        validity, boxes = load_conf_v1(conf, filename)
        return validity, version, boxes, {}
    elif version == 2:
        validity, boxes, cm = load_conf_v2(conf, filename)
        return validity, version, boxes, cm
    else:
        logger.error(f"CANCELED: wrong version '{version}' (!= 1, 2)")
        return False, version, [], {}


def check_link_validity(name: str, box: Quel1Box, args: argparse.Namespace) -> bool:
    status = box.reconnect(
        background_noise_threshold=args.background_noise_threshold,
        ignore_crc_error_of_mxfe=box.css.get_all_mxfes(),
        ignore_invalid_linkstatus=True,
    )
    if not all(status.values()):
        logger.warning("no valid link status, it is subject to be linked up")
        return False
    return True


def get_crc_error_count(box: Quel1Box) -> Dict[int, List[int]]:
    cntr: Dict[int, List[int]] = {}
    for mxfe_idx in box.css.get_all_mxfes():
        cntr[mxfe_idx] = box.css.get_crc_error_counts(mxfe_idx)
    return cntr


def diff_crc_error_count(
    cnts0: Dict[int, List[int]], cnts1: Dict[int, List[int]], ignore_crc_error_of_mxfe: Collection[int]
) -> bool:
    for mxfe_idx in cnts0:
        for lane_idx, cnt in enumerate(cnts0[mxfe_idx]):
            if cnt != cnts1[mxfe_idx][lane_idx]:
                if mxfe_idx in ignore_crc_error_of_mxfe:
                    logger.warning(
                        f"a new crc error is detected at the {lane_idx}-th lane of the mxfe-#{mxfe_idx}: "
                        f"{cnt} -> {cnts1[mxfe_idx][lane_idx]}, but ignore it."
                    )
                    return True
                else:
                    logger.warning(
                        f"a new crc error is detected at the {lane_idx}-th lane of the mxfe-#{mxfe_idx}: "
                        f"{cnt} -> {cnts1[mxfe_idx][lane_idx]}"
                    )
                    return False

    return True


def check_crc_error(name: str, box: Quel1Box, duration: int) -> bool:
    # Notes: all the mxfe should be linked up in advance.
    step = 0.5
    t = t0 = time.perf_counter()
    e0 = get_crc_error_count(box)
    while t < t0 + duration:
        time.sleep(step)
        t = time.perf_counter()
        e1 = get_crc_error_count(box)
        if not diff_crc_error_count(e0, e1, box.options["ignore_crc_error_of_mxfe"]):
            return False
        step = min(10.0, step * 2)

    # Notes: clear CRC error flag if no CRC error is detected.
    for mxfe_idx in box.css.get_all_mxfes():
        box.css.clear_crc_error(mxfe_idx)
    return True


# Notes: no leakage of exceptions is allowed. this function is submitted to pool.
#        all the possible exceptions are handled within function for better diagnosis.
def linkup_a_box(name: str, box: Quel1Box, args: argparse.Namespace) -> Union[bool, None]:
    threading.current_thread().name = name  # Notes: thread name is printed in log messages.

    if args.force:
        total_status = False
    else:
        try:
            total_status = check_link_validity(name, box, args)
        except Exception as e:
            logger.error(f"abort linking up due to {e}")
            return None

    for retry_idx in range(NUM_LINKUP_RETRY):
        if not total_status:
            # Notes: no log messages are shown during the link-up process which can last 30 seconds or so.
            logger.info("link-up is in progress")
            try:
                status = box.relinkup(
                    hard_reset_wss=args.hard_reset_wss,
                    background_noise_threshold=args.background_noise_threshold,
                )
                total_status = all(status.values())
                if not total_status:
                    continue
            except Exception as e:
                logger.error(e)
                # Notes: allow to retry anyway.
                continue

        # Notes: here total_status must be True
        logger.info(f"observing the link stability for {args.check_duration} seconds")
        try:
            if check_crc_error(name, box, args.check_duration):
                logger.info("no CRC error is detected")
                break
            else:
                total_status = False
        except Exception as e:
            logger.error(e)
            # Notes: total_status is not changed because no new information is available.
            continue
    else:
        return False

    return True


# Notes: no leakage of exceptions is allowed. this function is submitted to pool.
#        all the possible exceptions are handled within function for better diagnosis.
def reconnect_a_box(name: str, box: Quel1Box, args: argparse.Namespace) -> Union[bool, None]:
    threading.current_thread().name = name  # Notes: thread name is printed in log messages.

    try:
        total_status = check_link_validity(name, box, args)  # Notes: don't care of CRC error flags.
        if not total_status:
            logger.info("link is NOT ready")
    except Exception as e:
        logger.error(f"abort reconnecting due to {e}")
        return None

    return total_status


def _suppressing_noisy_loggers() -> None:
    logging.getLogger("quel_ic_config.ad9082").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.lmx2594").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1_config_subsystem_common").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1_config_subsystem_tempctrl").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1_config_loader").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.exstickge_coap_client").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1se_config_subsystem").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1se_riken8_config_subsystem").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.quel1se_fujitsu11_config_subsystem").setLevel(logging.WARNING)
    logging.getLogger("e7awghal.versionchecker").setLevel(logging.WARNING)
    logging.getLogger("quel_ic_config.exstickge_sock_client").disabled = True
    logging.getLogger("parallel_linkup").disabled = True
    logging.getLogger("quel1_parallel_linkup").disabled = True
    logging.getLogger("coap").disabled = True
    logging.getLogger("flufl.lock").disabled = True


def parallel_linkup_main() -> int:
    logging.basicConfig(
        level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: ({threadName}) {message}", style="{"
    )
    parser = argparse.ArgumentParser(description="parallelized link-up of QuEL-1 series control box")
    parser.add_argument("--conf", type=pathlib.Path, required=True, help="path to a configuration file")
    parser.add_argument("--ignore_unavailable", action="store_true", help="just ignoring unavailable boxes")
    parser.add_argument("--force", action="store_true", help="force to re-linkup all the boxes")
    parser.add_argument(
        "--check_duration", type=int, default=120, help="duration in seconds for checking the stability of the link"
    )
    parser.add_argument(
        "--background_noise_threshold",
        type=float,
        default=None,
        help="maximum allowable background noise amplitude of the ADCs at relinkup and at reconnect",
    )
    parser.add_argument(
        "--hard_reset_wss",
        action="store_true",
        default=False,
        help="hard reset AWG units and CAP units to clear hardware error flags, should avoid using it when unnecessary",
    )
    parser.add_argument("--verbose", action="store_true", help="show verbose log")
    args = parser.parse_args()

    if not args.verbose:
        _suppressing_noisy_loggers()

    # Notes: phase-1 loading configurations of boxes
    validity, version, box_confs, cm_conf = load_conf(pathlib.Path(args.conf))
    if not validity:
        return 1

    # Notes: phase-2 creating box objects
    pool = ThreadPoolExecutor(max_workers=len(box_confs))

    futures_box: list[Future[tuple[str, Union[Quel1Box, str]]]] = [
        pool.submit(load_box, box_conf) for box_conf in box_confs
    ]
    boxes: dict[str, Union[Quel1Box, str]] = {}
    for future_box in futures_box:
        box_name, box_obj = future_box.result()
        boxes[box_name] = box_obj

    # Notes: phase-3 linking up the boxes
    futures_status: Dict[str, Union[Future[Union[bool, None]], None]] = {}
    for name, obj in boxes.items():
        if isinstance(obj, Quel1Box):
            futures_status[name] = pool.submit(linkup_a_box, name, obj, args)
        else:
            # Notes: to show the final results in the same order as the given YAML file.
            futures_status[name] = None

    # Notes: taking a barrier sync to show the results after finishing all the threads
    num_failed: int = 0
    num_unavailable: int = 0
    results: Dict[str, str] = {}
    for name, future in futures_status.items():
        if future is not None:
            if future.result():
                results[name] = "ready"
            else:
                results[name] = "failed"
                num_failed += 1
        else:
            results[name] = cast(str, boxes[name])  # Notes: the content of boxes[name] is definitely str.
            num_unavailable += 1

    # Notes: phase-4 showing the results
    print("----------")
    for name, result in results.items():
        if result == "ready":
            print(f"{name} is ready to use")
        elif result == "failed":
            print(f"{name} fails to link up")
        else:
            print(f"{name} is unavailable due to '{result}'")
    print("----------")

    del boxes

    is_success: bool = False
    if num_failed == 0:
        if num_unavailable == 0:
            is_success = True
            logger.info(f"SUCCESS: all the boxes described in in '{args.conf}' are ready")
        elif args.ignore_unavailable:
            is_success = True
            logger.info(f"SUCCESS: but some boxes described in '{args.conf}' are unavailable")
        else:
            logger.error(f"FAILED: some boxes described in '{args.conf}' are unavailable")
    else:
        logger.error(f"FAILED: some boxes described in '{args.conf}' are not ready")

    return 0 if is_success else 1


def _calc_counter_delta(status: dict[str, tuple[bool, bool, int]]) -> tuple[bool, int, int, bool]:
    crl = []
    for name, (av, ls, tc) in status.items():
        if av and ls:
            cr = tc % 2000
            crl.append(cr)

    if len(crl) == 0:
        return False, 0, 0, False

    shifted: bool = False
    mincr = min(crl)
    maxcr = max(crl)
    if maxcr - mincr >= 1000:
        crl = [(cr - 1000) % 2000 + 1000 for cr in crl]
        mincr = min(crl)
        maxcr = max(crl)
        if maxcr - mincr < 1000:
            shifted = True
    return True, int(maxcr - mincr), mincr, shifted


def _print_counter(status: dict[str, tuple[bool, bool, int]], min_cntr: int, shifted: bool):
    print(f"base_sysref_offset: {min_cntr}")
    print()

    for name, (av, ls, tc) in status.items():
        if av and ls:
            if shifted:
                cr = (tc - 1000) % 2000 + 1000
            else:
                cr = tc % 2000
            print(f"{name}: {cr - min_cntr}")

    for name, (av, ls, tc) in status.items():
        if not av:
            print(f"{name}: not available, ignored")
        elif not ls:
            print(f"{name}: not ready to use, ignored")


def _print_sync_judge(delta: int, max_delta: int) -> bool:
    if delta >= 1000:
        logger.error(f"FAILED SYNCHRONIZATION, delta of the time counters is over 999 counts (> {max_delta:d} counts)")
    else:
        if delta <= max_delta:
            logger.info(
                "SUCCESSFUL SYNCHRONIZATION, "
                f"delta of the time counters is {delta:d} counts (<= {max_delta:d} counts)"
            )
            return True
        else:
            logger.error(
                f"FAILED SYNCHRONIZATION, delta of the time counters is {delta:d} counts (> {max_delta:d} counts)"
            )

    return False


def _sync_main(check_only: bool) -> int:
    logging.basicConfig(
        level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: ({threadName}) {message}", style="{"
    )
    parser = argparse.ArgumentParser(description="parallelized link-up of QuEL-1 series control box")
    parser.add_argument("--conf", type=pathlib.Path, required=True, help="path to a configuration file")
    parser.add_argument(
        "--background_noise_threshold",
        type=float,
        default=None,
        help="maximum allowable background noise amplitude of the ADCs at relinkup and at reconnect",
    )
    parser.add_argument("--ignore_unavailable", action="store_true", help="just ignore unavailable boxes exist")
    parser.add_argument(
        "--max_delta",
        type=int,
        default=DEFAULT_MAX_ALLOWABLE_COUNTERDELTA,
        help="worst allowable synchronization error",
    )
    parser.add_argument("--force", action="store_true", help="force to re-linkup all the boxes")
    parser.add_argument("--verbose", action="store_true", help="show verbose log")
    args = parser.parse_args()

    if not args.verbose:
        _suppressing_noisy_loggers()

    # Notes: phase-1 loading configurations of boxes
    validity, version, box_confs, cm_conf = load_conf(pathlib.Path(args.conf))
    if not validity:
        return 1

    # Notes: phase-2 creating box objects
    pool = ThreadPoolExecutor(max_workers=len(box_confs))

    futures_box: list[Future[tuple[str, Union[Quel1Box, str]]]] = [
        pool.submit(load_box, box_conf) for box_conf in box_confs
    ]
    boxes: dict[str, Union[Quel1Box, str]] = {}
    for future_box in futures_box:
        box_name, box_obj = future_box.result()
        boxes[box_name] = box_obj

    clock_master: Union[QuelClockMasterV1, None] = None
    if not check_only:
        if version >= 2:
            clock_master = QuelClockMasterV1(
                ipaddr=cm_conf["ipaddr"], boxes=[b for b in boxes.values() if isinstance(b, Quel1Box)]
            )
            if not clock_master.check_availability():
                logger.error(
                    f"CANCELED SYNCHRONIZATION due to the unavailability of the clockmaster {cm_conf['ipaddr']}"
                )
                return 1
        else:
            logger.error(
                f"CANCELED SYNCHRONIZATION because no clockmaster is specified in the config file '{args.conf}'"
            )
            return 1

    # Notes: phase-3 reconnecting to the boxes
    futures_status: Dict[str, Union[Future[Union[bool, None]], None]] = {}
    for name, obj in boxes.items():
        if isinstance(obj, Quel1Box):
            futures_status[name] = pool.submit(reconnect_a_box, name, obj, args)
        else:
            # Notes: to show the final results in the same order as the given YAML file.
            futures_status[name] = None

    # Notes: taking a barrier sync to show the results after finishing all the threads
    status: Dict[str, Tuple[bool, bool, int]] = {}
    ignored_box: list[str] = []
    for name, future in futures_status.items():
        if future is not None:
            box = cast(Quel1Box, boxes[name])
            if future.result():
                status[name] = True, True, int(box.get_latest_sysref_timecounter())
            else:
                status[name] = True, False, int(box.get_latest_sysref_timecounter())
                ignored_box.append(name)
        else:
            status[name] = False, False, 0
            ignored_box.append(name)

    # Notes: phase-4 showing the results
    if not args.ignore_unavailable and len(ignored_box) > 0:
        logger.error(f"CANCELED SYNCHRONIZATION because some boxes are not ready to use: {', '.join(ignored_box)}")
        return 1

    valid, delta, min_cntr, shifted = _calc_counter_delta(status)
    if not valid:
        logger.error("CANCELED SYNCHRONIZATION due to no available boxes")
        return 1

    # Notes: phase-5 kick
    if not check_only:
        if delta <= args.max_delta and not args.force:
            logger.info("all the boxes are synchronized well, skipping resynchronization")
        else:
            assert clock_master is not None  # TODO: replace assert with an appropriate mypy feature in future
            logger.info("conducting resynchronization")
            clock_master.sync_boxes()

            for name, (av, ls, tc) in status.items():
                if av:
                    box = cast(Quel1Box, boxes[name])
                    status[name] = (av, ls, int(box.get_latest_sysref_timecounter()))
            valid, delta, min_cntr, shifted = _calc_counter_delta(status)

    # Notes: phase-6 reporting
    print("----------")
    _print_counter(status, min_cntr, shifted)
    print("----------")
    sync_judge = _print_sync_judge(delta, args.max_delta)

    del boxes
    return 0 if sync_judge else 1


def syncstatus_main() -> int:
    return _sync_main(check_only=True)


def sync_main() -> int:
    return _sync_main(check_only=False)


if __name__ == "__main__":
    import sys

    sys.exit(parallel_linkup_main())



import json
import logging
import os.path as osp
import sys
import time
from abc import ABCMeta, abstractmethod
from argparse import ArgumentParser, Namespace
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, Final, Optional, Set, Tuple, Type, Union

import numpy as np
import numpy.typing as npt
import polars as pl
from e7awghal import AwgParam, CapParam, CapSection, WaveChunk
from quel_cmod_scripting import QuelCmod
from quel_pyxsdb import get_jtagterminal_port

from quel_ic_config import Quel1Box, Quel1BoxType, Quel1PortType, Quel1seAnyConfigSubsystem, Quel1Thermistor
from quel_ic_config_cli.stability_data_models import (
    QUEL1_ACTUATORS,
    QUEL1_THERMISTORS,
    QuelDataModel,
    TempCtrlStatus,
    WaveStatistics,
    get_tempctrlstatus_from_boxtype,
)
from quel_ic_config_utils import add_common_arguments, add_common_workaround_arguments, complete_ipaddrs

logger = logging.getLogger(__name__)

_DEFAULT_CONFIG_PATH: Final[Path] = Path(osp.dirname(__file__)) / "settings"


class StabilityMeasurement(metaclass=ABCMeta):
    @abstractmethod
    def get_stability_info(self, time_from_start: float) -> list[QuelDataModel]: ...


# Notes: I guess StabilityMeasurement is too abstract, keep it before its necessity is clarified.
class TemperatureMeasurement(StabilityMeasurement):
    def __init__(self, boxtype: Quel1BoxType):
        self._tcscls: Type[TempCtrlStatus] = get_tempctrlstatus_from_boxtype(boxtype)


class Quel1TemperatureMeasurement(TemperatureMeasurement):
    DEFAULT_XSDB_PORT_FOR_CMOD: Final[int] = 36335
    DEFAULT_HWSVR_PORT_FOR_CMOD: Final[int] = 6121
    DEFAULT_PERIOD: Final[float] = 10.0

    _THERMISTORS: dict[int, Quel1Thermistor] = QUEL1_THERMISTORS
    _ACTUATORS: dict[str, tuple[str, int]] = QUEL1_ACTUATORS

    def __init__(self, cmod: QuelCmod, boxtype: Quel1BoxType):
        super().__init__(boxtype)
        self._cmod = cmod
        self._ver = cmod.ver()

    def _measure(self) -> dict[str, float]:
        for _ in range(3):
            t0 = self._cmod.thall()
            if t0 is None:
                continue
            return self._convert(t0)
        else:
            raise RuntimeError("failed to acquire temperature repeatedly")

    def _plstat(self) -> npt.NDArray[np.int32]:
        p = self._cmod.plstat()
        if p is None:
            raise RuntimeError("failed to acquire peltier drive values")
        return p

    def _convert(self, adc: npt.NDArray[np.int32]) -> dict[str, float]:
        return {th.name: round(th.convert(adc[idx]), 2) for idx, th in self._THERMISTORS.items()}

    def _make_tempctrl_stat(
        self, time_from_start: float, temperatures: dict[str, float], actuators: npt.NDArray[np.int32]
    ) -> list[QuelDataModel]:
        tempctrl_dicts: list[QuelDataModel] = []
        for name, temp in temperatures.items():
            a_type, a_idx = self._ACTUATORS.get(name, (None, None))
            tempctrl_model = self._tcscls(
                time_from_start=time_from_start,
                location_name=name,
                temperature=temp,
                actuator_type=a_type,
                actuator_val=float(actuators[a_idx]) if a_idx is not None else None,
            )
            tempctrl_dicts.append(tempctrl_model)

        return tempctrl_dicts

    def get_stability_info(self, time_from_start: float) -> list[QuelDataModel]:
        t0 = time.perf_counter()
        temps = self._measure()
        plstat = self._plstat()
        while time.perf_counter() - t0 < self.DEFAULT_PERIOD:
            time.sleep(0.25)
        return self._make_tempctrl_stat(time_from_start, temps, plstat)


class Quel1seTemperatureMeasurement(TemperatureMeasurement):
    def __init__(self, css: Quel1seAnyConfigSubsystem):
        super().__init__(css.boxtype)
        self._css: Quel1seAnyConfigSubsystem = css

    def _make_tempctrl_stat(
        self,
        time_from_start: float,
        temperatures: dict[str, float],
        actuators: dict[str, dict[str, float]],
    ) -> list[QuelDataModel]:
        tempmodels: list[QuelDataModel] = []

        for loc_name, temp in temperatures.items():
            if loc_name in actuators["fan"]:
                tm = self._tcscls(
                    time_from_start=time_from_start,
                    location_name=loc_name,
                    temperature=temp,
                    actuator_type="fan",
                    actuator_val=actuators["fan"][loc_name],
                )
            elif loc_name in actuators["heater"]:
                tm = self._tcscls(
                    time_from_start=time_from_start,
                    location_name=loc_name,
                    temperature=temp,
                    actuator_type="heater",
                    actuator_val=actuators["heater"][loc_name],
                )
            else:
                tm = self._tcscls(
                    time_from_start=time_from_start,
                    location_name=loc_name,
                    temperature=temp,
                    actuator_type=None,
                    actuator_val=None,
                )
            tempmodels.append(tm)

        return tempmodels

    def get_stability_info(self, time_from_start: float) -> list[QuelDataModel]:
        temperatures: dict[str, float] = self._css.get_tempctrl_temperature().result()
        for mxfe_idx in range(2):
            mxfe_temp_max, mxfe_temp_min = self._css.get_mxfe_temperature_range(mxfe_idx)
            temperatures[f"mxfe{mxfe_idx}_max"] = mxfe_temp_max
            temperatures[f"mxfe{mxfe_idx}_min"] = mxfe_temp_min
        actuators = self._css.get_tempctrl_actuator_output()
        return self._make_tempctrl_stat(time_from_start, temperatures, actuators)


class WaveStabilityMeasurement(StabilityMeasurement):
    _DEFAULT_NUM_SAMPLES_IN_EPOCH: Final[int] = 50000

    def __init__(
        self,
        box: Quel1Box,
        target_output_ports: Set[Quel1PortType],
    ):
        self._box = box
        self._wvcls: Type[WaveStatistics] = WaveStatistics
        self._inout_combinations = self._valid_inout_combinations(target_output_ports)
        self._init_awg_and_cap()

    def _is_matched_inout_frequency(self, in_port_dump: Dict[str, Any], out_port_dump: Dict[str, Any]) -> bool:
        # up-convert output frequency with LO and NCO frequency for channel 0
        if_freq = out_port_dump["channels"][0]["fnco_freq"] + out_port_dump["cnco_freq"]
        if "lo_freq" in out_port_dump.keys():
            # analog mixing
            out_freq = (
                if_freq + out_port_dump["lo_freq"]
                if out_port_dump["sideband"] == "U"
                else out_port_dump["lo_freq"] - if_freq
            )
        else:
            out_freq = if_freq

        # analog down-conversion with LO
        analog_dc_freq = abs(out_freq - in_port_dump["lo_freq"])
        # after digital down-conversion, check if the signal is DC
        if analog_dc_freq - in_port_dump["runits"][0]["fnco_freq"] - in_port_dump["cnco_freq"] == 0.0:
            return True

        return False

    def _valid_inout_combinations(
        self, target_output_ports: Set[Quel1PortType]
    ) -> Set[Tuple[Quel1PortType, Quel1PortType]]:
        boxdump = self._box.dump_box()["ports"]
        valid_combinations: Set[Tuple[Quel1PortType, Quel1PortType]] = set()
        for input_port in self._box.get_input_ports():
            for output_port in self._box.get_loopbacks_of_port(input_port):
                if output_port not in target_output_ports:
                    continue
                if self._is_matched_inout_frequency(boxdump[input_port], boxdump[output_port]):
                    valid_combinations.add((input_port, output_port))
        logger.info(f"valid input-output port combinations: {valid_combinations}")
        return valid_combinations

    def _init_awg_and_cap(self):
        cw_iq = np.zeros(64, dtype=np.complex64)
        cw_iq[:] = 32767.0 + 0.0j
        ap = AwgParam(num_wait_word=0, num_repeat=1024 + 32)
        ap.chunks.append(WaveChunk(name_of_wavedata="cw32767", num_blank_word=0, num_repeat=1))
        for port in self._box.get_output_ports():
            if port in {ports[1] for ports in self._inout_combinations}:
                self._box.register_wavedata(port, 0, "cw32767", cw_iq)
                self._box.config_channel(port, 0, awg_param=ap)

        cp = CapParam(num_repeat=1)
        cp.sections.append(CapSection(name="s0", num_capture_word=(16384 + 512)))
        for port in {ports[0] for ports in self._inout_combinations}:
            self._box.config_runit(port, 0, capture_param=cp)

    def phase_stat(
        self,
        time_from_start: float,
        input_port: Quel1PortType,
        output_port: Quel1PortType,
        iq: npt.NDArray[np.complex64],
    ) -> WaveStatistics:
        num_samples = self._DEFAULT_NUM_SAMPLES_IN_EPOCH
        if len(iq) < num_samples:
            logger.info(f"processing {len(iq)} samples")
        else:
            iq = iq[:num_samples]

        pwr: npt.NDArray[np.float64] = np.abs(iq)
        angle: npt.NDArray[np.float64] = np.angle(iq)
        # Notes: angle changes from pi --> -pi suddenly.
        if max(angle) >= 3.0 and min(angle) < -3.0:
            angle = (angle + 2 * np.pi) % (2 * np.pi)  # the max value is (-3.0 + 2*pi) % (2*pi) = pi + (pi-3.0) > pi

        pwr_mean: float = float(np.mean(pwr))
        pwr_std: float = float(np.sqrt(np.var(pwr)))
        agl_mean: float = float(np.mean(angle)) * 180.0 / np.pi
        agl_std: float = float(np.sqrt(np.var(angle))) * 180.0 / np.pi
        agl_deltamax: float = float((np.max(angle) - np.min(angle))) * 180.0 / np.pi
        model = self._wvcls(
            time_from_start=time_from_start,
            input_port=str(input_port),
            output_port=str(output_port),
            power_mean=round(pwr_mean, 3),
            power_std=round(pwr_std, 3),
            angle_mean=round(agl_mean, 3),
            angle_std=round(agl_std, 3),
            angle_deltamax=round(agl_deltamax, 3),
        )
        return model

    def get_stability_info(self, time_from_start: float) -> list[QuelDataModel]:
        wavestability_list: list[QuelDataModel] = []
        for input_port, output_port in self._inout_combinations:
            c_task, g_task = self._box.start_capture_by_awg_trigger(
                {(input_port, 0)},
                {(output_port, 0)},
            )
            assert g_task.result() is None
            rdrs0 = c_task.result()
            data0 = rdrs0[input_port, 0].as_wave_dict()
            iq = data0["s0"][0]
            model = self.phase_stat(time_from_start, input_port, output_port, iq[1024:-1024])
            wavestability_list.append(model)
        return wavestability_list


class DataHandler:
    def __init__(
        self,
        name: str,
        label: str,
        measurement: StabilityMeasurement,
        start_time: float,
        out_dir_path: Path,
        interval: Optional[float] = None,
    ):
        self._name: str = name
        self._label: str = label
        self._measurement: StabilityMeasurement = measurement
        self._header_flag: bool = True
        self._start_time: float = start_time
        self._interval: Union[float, None] = interval
        self._filepath: Path = self.make_csv_path(out_dir_path)

    def make_csv_path(self, out_dir_path: Path) -> Path:
        out_subdir_path = out_dir_path / Path(self._name)
        out_subdir_path.mkdir(parents=True, exist_ok=True)
        if not out_subdir_path.exists():
            raise RuntimeError(f"failed to create a directory '{str(out_subdir_path)}")
        return out_subdir_path / Path(f"{self._name}_{self._label}.csv")

    def get_and_write_data(self):
        time_from_start: float = time.perf_counter() - self._start_time
        model_list = self._measurement.get_stability_info(time_from_start)
        model_dumps = [model.model_dump() for model in model_list]
        for m in model_dumps:
            logger.info(json.dumps(m))
        df_stability = pl.DataFrame(model_dumps)
        with open(self._filepath, mode="ab") as f:
            df_stability.write_csv(f, include_header=self._header_flag)
        if self._header_flag:
            self._header_flag = False
        if isinstance(self._measurement, Quel1TemperatureMeasurement) and (self._interval is not None):
            time.sleep(self._interval)


def parse_target_ports(optstr: str) -> set[Quel1PortType]:
    tokens = [t.strip() for t in optstr.split(",")]
    ctx: list[str] = []
    outports = set()
    for cur in tokens:
        if len(cur) == 0:
            raise ValueError
        if cur.startswith("("):
            if len(ctx) != 0:
                raise ValueError
            else:
                ctx.append(cur)
        elif cur.endswith(")"):
            if len(ctx) != 1:
                raise ValueError
            else:
                outport: Quel1PortType = (int(ctx[0][1:]), int(cur[:-1]))
                ctx.clear()
                if outport in outports:
                    raise ValueError
                else:
                    outports.add(outport)
        else:
            if len(ctx) != 0:
                raise ValueError
            else:
                outport = int(cur)
                if outport in outports:
                    raise ValueError
                else:
                    outports.add(outport)
    return outports


def load_configs(box: Quel1Box, json_path: Optional[Path] = None) -> None:
    boxtype = Quel1BoxType.fromstr(box.boxtype)
    if boxtype in {Quel1BoxType.QuEL1SE_RIKEN8, Quel1BoxType.QuEL1SE_RIKEN8DBG}:
        json_name: str = "wave_stability_quel1se-riken8.json"
    elif boxtype in {Quel1BoxType.QuEL1SE_FUJITSU11_TypeA, Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeA}:
        json_name = "wave_stability_quel1-a.json"
    elif boxtype in {Quel1BoxType.QuEL1SE_FUJITSU11_TypeB, Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeB}:
        json_name = "wave_stability_quel1-b.json"
    elif boxtype == Quel1BoxType.QuEL1_TypeA:
        json_name = "wave_stability_quel1-a.json"
    elif boxtype == Quel1BoxType.QuEL1_TypeB:
        json_name = "wave_stability_quel1-b.json"
    else:
        raise ValueError(f"{boxtype} is not supported yet")

    json_path = json_path or _DEFAULT_CONFIG_PATH / json_name
    if json_path.exists():
        box.config_box_from_jsonfile(json_path)
    else:
        raise ValueError(f"'{json_path}' does not exist.")


def dump_configs(
    box: Quel1Box,
    out_dir_path: Path,
    title: str,
):
    out_dir_path_configs = out_dir_path / Path("configs")
    out_dir_path_configs.mkdir(parents=True, exist_ok=True)
    if not out_dir_path_configs.exists():
        raise RuntimeError(f"failed to create a directory '{str(out_dir_path_configs)}'")

    json_file_path = out_dir_path_configs / Path(f"configs_{title}.json")
    box.dump_box_to_jsonfile(json_file_path)


def cli_args() -> Namespace:
    logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    logging.getLogger("quel_ic_config_utils.quel1_wave_subsystem").setLevel(logging.WARNING)

    parser = ArgumentParser(description="a phase fluctuation measurement tool")
    add_common_arguments(parser, use_config_root=False, use_config_options=False)
    add_common_workaround_arguments(parser, use_ignore_crc_error_of_mxfe=True)
    parser.add_argument(
        "--cmod_host",
        type=str,
        default="localhost",
        help="ip address of host PC of Cmod USB",
    )
    parser.add_argument(
        "--xsdb_port",
        type=int,
        default=Quel1TemperatureMeasurement.DEFAULT_XSDB_PORT_FOR_CMOD,
        help="port of xsdb managing Cmod USB",
    )
    parser.add_argument(
        "--hwsvr_port",
        type=int,
        default=Quel1TemperatureMeasurement.DEFAULT_HWSVR_PORT_FOR_CMOD,
        help="port of hw_server managing Cmod USB",
    )
    parser.add_argument("--cmod_jtag", type=str, default="", help="jtag id of the Cmod USB adapter")
    parser.add_argument("--conf", type=Path, default=None, help="json file for configuration")
    parser.add_argument("--duration", type=int, default=30, help="measurement duration in second")
    parser.add_argument("--outdir", type=Path, required=True, help="output file directory")
    parser.add_argument("--skip_configs", action="store_true", default=False, help="skip configurations")
    parser.add_argument("--boxname", type=str, default=None, help="custom box name")
    parser.add_argument(
        "--outports",
        type=parse_target_ports,
        default=None,
        help="target output ports with ',' seperation. ex. '1,(1,1),6'. ",
    )
    args = parser.parse_args()

    complete_ipaddrs(args)

    return args


def cli_init(args: Namespace):
    box: Quel1Box = Quel1Box.create(
        ipaddr_wss=str(args.ipaddr_wss),
        ipaddr_sss=str(args.ipaddr_sss),
        ipaddr_css=str(args.ipaddr_css),
        boxtype=args.boxtype,
    )
    status = box.reconnect()

    target_output_ports: set[Quel1PortType] = args.outports or box.get_output_ports()
    required_mxfe: set[int] = set()
    for outport in target_output_ports:
        # TODO: this part is tricky, consider better ways...
        gr, ln = box._convert_output_port(outport)
        mxfe_idx, _ = box.css.get_dac_idx(gr, ln)
        required_mxfe.add(mxfe_idx)

    for mxfe_idx in required_mxfe:
        if not status[mxfe_idx]:
            raise RuntimeError(f"mxfe-#{mxfe_idx} of box {str(args.ipaddr_wss)} is not linked up properly")

    if args.boxtype.is_quel1se():
        if isinstance(box.css, Quel1seAnyConfigSubsystem):
            tmeas: Union[TemperatureMeasurement, None] = Quel1seTemperatureMeasurement(box.css)
        else:
            raise AssertionError("never happens")
    else:
        if args.cmod_jtag != "":
            cmod_port = get_jtagterminal_port(
                adapter_id=args.cmod_jtag, host=args.cmod_host, xsdb_port=args.xsdb_port, hwsvr_port=args.hwsvr_port
            )
            cmod = QuelCmod(host=args.cmod_host, port=cmod_port)
            tmeas = Quel1TemperatureMeasurement(cmod, args.boxtype)
        else:
            tmeas = None
            logger.warning("ONLY PHASE MEASUREMENT will be conducted since jtag adapter id is not specified")

    if args.skip_configs is False:
        load_configs(box, args.conf)
    pmeas = WaveStabilityMeasurement(box, target_output_ports)
    return box, tmeas, pmeas


def cli_body(args: Namespace, box: Quel1Box, tmeas: Optional[StabilityMeasurement], pmeas: StabilityMeasurement):
    output_file_label = f"{str(args.ipaddr_wss).replace('.','_')}_{time.strftime('%Y%m%d%H%M%S')}"
    if args.boxname is not None:
        output_file_label = args.boxname + "_" + output_file_label
    dump_configs(box, args.outdir, output_file_label)
    start_time = time.perf_counter()
    data_handlers: list[DataHandler] = []
    data_handlers.append(DataHandler("wavestability", output_file_label, pmeas, start_time, args.outdir))
    if tmeas is not None:
        data_handlers.append(DataHandler("tempctrl", output_file_label, tmeas, start_time, args.outdir))

    executor = ThreadPoolExecutor(max_workers=len(data_handlers))
    futures: list[Future[None]] = []
    for data_handler in data_handlers:
        futures.append(executor.submit(data_handler.get_and_write_data))

    while (time.perf_counter() - start_time) < args.duration:
        for idx, future in enumerate(futures):
            if future.done():
                future.result()
                futures[idx] = executor.submit(data_handlers[idx].get_and_write_data)
        time.sleep(0.01)

    for future in futures:
        future.result()


def cli_main():
    box: Union[Quel1Box, None] = None
    try:
        args = cli_args()
        box, tmeas, pmeas = cli_init(args)
        if box is None:
            raise AssertionError("never happens")
        cli_body(args, box, tmeas, pmeas)
    except Exception as e:
        logger.error(e)
        sys.exit(1)
    finally:
        if box:
            del box

    sys.exit(0)


if __name__ == "__main__":
    args0 = cli_args()
    box0, tmeas0, pmeas0 = cli_init(args0)
    cli_body(args0, box0, tmeas0, pmeas0)



import csv
import glob
import logging
import os
import re
import sys
import time
from abc import ABCMeta, abstractmethod
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, Final, List, Literal, Optional, Set, Tuple, Type

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.widgets import CheckButtons
from pandas import Series

from quel_ic_config import QUEL1_BOXTYPE_ALIAS, Quel1BoxType
from quel_ic_config_cli.stability_data_models import QuelDataModel, WaveStatistics, get_tempctrlstatus_from_boxtype
from quel_ic_config_utils.common_arguments import parse_boxtype

logger = logging.getLogger()
UPDATE_INTERVAL: Final[int] = 20000


class AbstractGraph(metaclass=ABCMeta):
    WARN_TIME_THRESHOLD: Final[float] = 60.0

    def __init__(
        self,
        csv_path: Path,
        time_window: Tuple[Optional[float], Optional[float]],
        boxtype: Optional[Quel1BoxType] = None,
    ):
        self._time_window: Tuple[Optional[float], Optional[float]] = time_window
        self._latest_time = 0.0
        self._file = open(csv_path, "r")
        self._csv_reader = csv.DictReader(self._file)
        self._schema = self._get_schema(boxtype)

    def _get_schema(self, boxtype: Optional[Quel1BoxType]) -> Type[QuelDataModel]:
        if isinstance(self, WaveGraph):
            return WaveStatistics
        elif isinstance(self, TempCtrlGraph):
            if boxtype is not None:
                return get_tempctrlstatus_from_boxtype(boxtype)
            else:
                raise ValueError("boxtype is required to get tempctrl schema")
        else:
            raise ValueError(f"{type(self)} is not a recognized subclass of AbstractGraph")

    def _get_new_lines(self) -> Optional[pd.DataFrame]:
        try:
            new_lines: List[QuelDataModel] = []
            for row in self._csv_reader:
                # Convert empty strings to None
                processed_row = {
                    key: (None if value == "" else value) for key, value in row.items() if key != "thermistor_id"
                }
                new_lines.append(self._schema.model_validate(processed_row))
            if new_lines:
                self._latest_time = time.time()  # Store the current time when data comes
                return pd.DataFrame([quelmodel.model_dump() for quelmodel in new_lines])
            else:
                elapsed_time = time.time() - self._latest_time
                if elapsed_time >= self.WARN_TIME_THRESHOLD:
                    logger.warning(f"No data received for {int(elapsed_time)} seconds since the last valid data.")
                return None

        except OSError as e:
            logger.error(f"An unexpected error occurred while accessing the file: {e}")
            self._file.close()
            sys.exit(1)

    @abstractmethod
    def _initialize(self):
        pass

    @abstractmethod
    def update_graph(self, frame: int):  # need argument "frame" for FuncAnimation
        pass

    @property
    @abstractmethod
    def fig(self) -> Figure:
        pass

    def _update_axes(self, axs, fig):
        for ax in axs:
            ax.relim(visible_only=True)
            ax.set_autoscalex_on(True)
            ax.autoscale_view()
            xmin, xmax = ax.get_xlim()
            # when _time_window[i] is None, set autoscaled values
            xmin = self._time_window[0] if self._time_window[0] is not None else xmin
            xmax = self._time_window[1] if self._time_window[1] is not None else xmax
            ax.set_xlim(xmin, xmax)
        fig.tight_layout()
        fig.canvas.draw_idle()


class WaveGraph(AbstractGraph):
    _COLUMNS_HEADER: List[str] = [
        "power_mean",
        "power_std",
        "angle_mean",
        "angle_std",
        "angle_deltamax",
    ]
    _COLUMNS_UNIT: List[str] = [
        "ADC Counts",
        "ADC Counts",
        "DEGREE",
        "DEGREE",
        "DEGREE",
    ]

    def __init__(self, csv_path: Path, time_window: Tuple[Optional[float], Optional[float]]):
        super().__init__(csv_path, time_window)
        self._fig, self._axs = plt.subplots(2, 3, figsize=(10, 7))
        fig_manager = self._fig.canvas.manager
        if fig_manager is not None:
            fig_manager.set_window_title("Waveform Stability")
        self._checkbuttons: Optional[CheckButtons] = None
        self._lines: Dict[str, List[Line2D]] = {}
        self._port_combination_labels: Set[str] = set()
        self._angle_mean_offsets: Dict[Tuple[str, str], float] = {}
        self._initialize()

    @property
    def fig(self) -> Figure:
        return self._fig

    def _set_graph_label(self):
        for idx, ax in enumerate(self._axs.flat):
            if idx < len(self._COLUMNS_HEADER):
                ax.set_ylabel(f"{self._COLUMNS_HEADER[idx]} ({self._COLUMNS_UNIT[idx]})")
                ax.set_xlabel("Time From Start (s)")
                ax.grid(axis="both")
            else:
                ax.axis("off")

    def _check_func(self, label: Optional[str]):
        if label is None:
            raise AssertionError("never happan")
        for line in self._lines[label]:
            line.set_visible(not line.get_visible())
        self._update_axes(self._axs.flat, self._fig)

    def _create_checkbuttons(self):
        labels = [label for label in self._lines.keys()]
        line_colors = []
        for label in labels:
            line = self._lines[label][0]
            line_colors.append(line.get_color())
        self._checkbuttons = CheckButtons(
            self._axs.flat[-1],
            labels=labels,
            actives=[True] * len(labels),
            label_props={"color": line_colors},
            frame_props={"edgecolor": line_colors},
            check_props={"facecolor": line_colors},
        )
        if self._checkbuttons is not None:
            self._checkbuttons.on_clicked(self._check_func)

    def _draw_lines(
        self,
        port_combination: Tuple[int, int],
        group: pd.DataFrame,
        option: Literal["initialize", "update"],
    ):
        label = f"Output {port_combination[0]} - Input {port_combination[1]}"
        if option == "initialize":
            self._port_combination_labels.add(label)
            self._lines[label] = []
            for idx, ax in enumerate(self._axs.flat):
                if idx < len(self._COLUMNS_HEADER):
                    (line,) = ax.plot(group["time_from_start"], group[self._COLUMNS_HEADER[idx]], label=label)
                    self._lines[label].append(line)
        elif option == "update":
            for idx, line in enumerate(self._lines[label]):
                x_data, y_data = line.get_data()
                x_data = np.concatenate([x_data, group["time_from_start"].values])
                y_data = np.concatenate([y_data, group[self._COLUMNS_HEADER[idx]].values])
                line.set_data(x_data, y_data)

    def _phase_offset_subtract(self, group: pd.DataFrame, offset: float) -> pd.DataFrame:
        group["angle_mean"] = group["angle_mean"] - offset

        # if offset corrected data is very large, it is regarded as a wrap-around
        # phase variations within +- 60 degrees are correctly displayed.
        group["angle_mean"] = group["angle_mean"].apply(
            lambda x: x - 360.0 if x > 60.0 else (x + 360.0 if x < -60.0 else x)
        )
        return group

    def _process_data(self, data_df: pd.DataFrame, option: Literal["initialize", "update"]):
        ports = data_df.groupby(["output_port", "input_port"])
        for port_combination, group in ports:
            if option == "initialize":
                self._angle_mean_offsets[port_combination] = group["angle_mean"].iloc[0]
            group = self._phase_offset_subtract(group, self._angle_mean_offsets[port_combination])
            self._draw_lines(port_combination, group, option)

    def _initialize(self):
        new_data_df: Optional[pd.DataFrame] = self._get_new_lines()
        if new_data_df is None:
            raise ValueError("no data found. the csv for wavestability may be empty")
        self._set_graph_label()
        self._process_data(new_data_df, "initialize")
        self._create_checkbuttons()
        self._update_axes(self._axs.flat, self._fig)

    def update_graph(self, frame: int):
        new_data_df: Optional[pd.DataFrame] = self._get_new_lines()
        if new_data_df is None:
            return
        self._process_data(new_data_df, "update")
        self._update_axes(self._axs.flat, self._fig)

    def print_wave_stability(self):
        start_time, end_time = (self._time_window[0] or 0.0, self._time_window[1] or 1e10)
        for label in self._port_combination_labels:
            print(f"\n{label}")
            print("-" * len(label))
            title_width = 15
            num_width = 10
            print(f"{'Title':<{title_width}} | {'Std':>{num_width}} | {'Mean':>{num_width}} | {'Ratio':>{num_width}}")
            print(f"{'-'*title_width}-+-{'-'*num_width}-+-{'-'*num_width}-+-{'-'*num_width}")
            for idx, line in enumerate(self._lines[label]):
                x_data, y_data = line.get_data()
                title = self._COLUMNS_HEADER[idx]
                x_data = np.array(x_data, dtype=float)
                y_data = np.array(y_data, dtype=float)
                mask = (x_data > start_time) & (x_data < end_time)
                std = np.std(y_data[mask])
                mean = np.mean(y_data[mask])
                if "angle_mean" in title:
                    ratio_str: str = "N/A"
                else:
                    ratio_str = f"{std/mean:.6f}"
                print(
                    f"{title:<{title_width}} | {std:>{num_width}.6f} |"
                    f" {mean:>{num_width}.6f} | {ratio_str:>{num_width}}"
                )
            print(f"{'-'*title_width}---{'-'*num_width}-+-{'-'*num_width}---{'-'*num_width}")


class TempCtrlGraph(AbstractGraph):
    _LOCATION_TYPE: Dict[str, int] = {}
    _CORRESPONDING_BOXTYPES: Set[Quel1BoxType] = set()

    def __init__(self, csv_path: Path, time_window: Tuple[Optional[float], Optional[float]], boxtype: Quel1BoxType):
        super().__init__(csv_path, time_window, boxtype)
        self._lines: Dict[str, Dict[str, List[Line2D]]] = {"temperature": {}, "actuator_val": {}}
        self._time_window: Tuple[Optional[float], Optional[float]] = time_window
        t_fig, t_axs = plt.subplots(2, 4, figsize=(12, 7))
        a_fig, a_axs = plt.subplots(2, 4, figsize=(12, 7))
        self._figs: Dict[str, Figure] = {"temperature": t_fig, "actuator_val": a_fig}
        self._axs: Dict[str, Tuple[Axes, ...]] = {"temperature": t_axs.flat, "actuator_val": a_axs.flat}
        for key, fig in self._figs.items():
            fig_manager = fig.canvas.manager
            if fig_manager is None:
                continue
            if key == "temperature":
                fig_manager.set_window_title("Temperatures")
            else:
                fig_manager.set_window_title("Actuator Values")
        self._initialize()

    @classmethod
    def is_corresponding_boxtype(cls, boxtype: Quel1BoxType) -> bool:
        return boxtype in cls._CORRESPONDING_BOXTYPES

    @property
    def fig(self) -> Figure:
        return self._figs["temperature"]

    def _set_graph_label(self):
        for key, axs in self._axs.items():
            for loc, idx in self._LOCATION_TYPE.items():
                axs[idx].set_title(loc)
                if key == "temperature":
                    axs[idx].set_ylabel(r"Temperature [$^\circ$C]")
                else:
                    axs[idx].set_ylabel("Actuator Strength")
                axs[idx].set_xlabel("Time From Start (s)")
                axs[idx].grid(axis="both")
                if axs[idx].get_legend_handles_labels()[0]:  # Check if there are any labels
                    axs[idx].legend(loc="upper left", fontsize=8)

    def _get_location_type(self, location: str) -> str:
        for location_type in self._LOCATION_TYPE.keys():
            if location_type not in location:
                continue
            else:
                return location_type
        else:
            raise ValueError(f"can not find location type for location: {location}")

    def _draw_lines(
        self,
        ax_key: Literal["temperature", "actuator_val"],
        location: str,
        data: Tuple["Series[float]", "Series[float]"],
        option: Literal["initialize", "update"],
    ):
        loc = location.replace("_", "")
        if option == "initialize":
            location_type = self._get_location_type(location)
            self._lines[ax_key][loc] = []
            label = loc.replace(location_type, "")
            (line,) = self._axs[ax_key][self._LOCATION_TYPE[location_type]].plot(data[0], data[1], label=label)
            self._lines[ax_key][loc].append(line)
        elif option == "update":
            loc = location.replace("_", "")
            old_x, old_y = self._lines[ax_key][loc][0].get_data()
            new_x = np.concatenate([old_x, data[0]])
            new_y = np.concatenate([old_y, data[1]])
            self._lines[ax_key][loc][0].set_data(new_x, new_y)

    def _process_data(self, data_df: pd.DataFrame, option: Literal["initialize", "update"]):
        for location, group in data_df.groupby("location_name"):
            assert isinstance(location, str)
            location_type = self._get_location_type(location)

            # there are too many thermistors in path selector boards.
            # only thermistors whose corresponding actuators exist are displayed
            if location_type in ("ps0", "ps1") and pd.isna(group["actuator_val"].values[0]):
                continue

            self._draw_lines("temperature", location, (group["time_from_start"], group["temperature"]), option)

            if pd.isna(group["actuator_val"].values[0]):
                continue
            self._draw_lines("actuator_val", location, (group["time_from_start"], group["actuator_val"]), option)

    def _initialize(self):
        new_data_df: Optional[pd.DataFrame] = self._get_new_lines()
        if new_data_df is None:
            raise ValueError("csv for tempctrl may be empty.")
        self._process_data(new_data_df, "initialize")
        self._set_graph_label()
        for key in ["temperature", "actuator_val"]:
            self._update_axes(self._axs[key], self._figs[key])

    def update_graph(self, frame: int):
        new_data_df: Optional[pd.DataFrame] = self._get_new_lines()
        if new_data_df is None:
            return
        self._process_data(new_data_df, "update")
        self._set_graph_label()
        for key in ["temperature", "actuator_val"]:
            self._update_axes(self._axs[key], self._figs[key])


class TempCtrlGraphQuel1(TempCtrlGraph):
    _LOCATION_TYPE: Dict[str, int] = {
        "ad9082": 0,
        "lmx2594": 1,
        "adrf6780": 2,
        "panel": 3,
        "rx": 4,
        "adclk": 5,
        "ps": 6,
    }
    _CORRESPONDING_BOXTYPES: Set[Quel1BoxType] = {
        Quel1BoxType.QuEL1_TypeA,
        Quel1BoxType.QuEL1_TypeB,
        Quel1BoxType.QuEL1_NEC,
        Quel1BoxType.QuEL1_NTT,
    }


class TempCtrlGraphQuel1seRiken8(TempCtrlGraph):
    _LOCATION_TYPE: Dict[str, int] = {
        "mxfe": 0,
        "lmx2594": 1,
        "adrf6780": 2,
        "panel": 3,
        "hmc8193": 4,
        "amp": 4,
        "ps0": 5,
        "ps1": 6,
    }
    _CORRESPONDING_BOXTYPES: Set[Quel1BoxType] = {
        Quel1BoxType.QuEL1SE_RIKEN8,
        Quel1BoxType.QuEL1SE_RIKEN8DBG,
    }


class TempCtrlGraphQuel1seFujitsu11(TempCtrlGraph):
    _LOCATION_TYPE: Dict[str, int] = {
        "mxfe": 0,
        "lmx2594": 1,
        "adrf6780": 2,
        "panel": 3,
        "rx": 4,
        "hmc8193": 5,
        "ps0": 6,
        "ps1": 7,
    }
    _CORRESPONDING_BOXTYPES: Set[Quel1BoxType] = {
        Quel1BoxType.QuEL1SE_FUJITSU11_TypeA,
        Quel1BoxType.QuEL1SE_FUJITSU11_TypeB,
        Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeA,
        Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeB,
    }


def validate_filename(filename):
    pattern = r"^wavestability_.*_\d{14}\.csv$"
    if not re.match(pattern, os.path.basename(filename)):
        raise ValueError(f"Invalid filename format: {filename}")
    return filename


def tempctrl_csv_path(wave_csv_file_str: str) -> Optional[Path]:

    # temperature file path is data_dir/tempctrl/tempctrl_*.csv
    # while wavestability file path is data_dir/wavestability/wavestability_*.csv
    temp_csv_file_path: Path = Path(wave_csv_file_str.replace("wavestability", "tempctrl"))
    if temp_csv_file_path.exists():
        return temp_csv_file_path

    # in case of specifying the relative path in the directory in data_dir/wavestability/
    temp_csv_file_path = Path("../tempctrl") / temp_csv_file_path
    if temp_csv_file_path.exists():
        return temp_csv_file_path

    logger.warning(f"Temperature control data file not found: {temp_csv_file_path}")
    return None


def get_csv_path(
    args: Namespace,
) -> Tuple[Path, Optional[Path]]:
    # Cases are now validated:
    # Case 1. args.wave_csv is specified; others are not (handled at the beginning)
    # Case 2. args.data_dir and args.ipaddr_wss are specified; others are not
    # Case 3. args.data_dir and args.boxname are specified; others are not

    # DATA LOCATION (OUTDIR is specified directory in quel1_stability_check command),
    # -wave data csv        : OUTDIR/wavestability
    # -temperature data csv :  OUTDIR/tempctrl

    # In Case 1, specify the wave stability CSV file directly (OUTDIR/wavestability/xxxx.csv),
    # and the temperature data CSV file is found automatically.
    # In Case 2 and Case 3, specify only OUTDIR and either the IP address or box name.
    # due to many candidates in this cases, the newest one is selected

    if args.wave_csv is not None:
        if args.data_dir is not None or args.ipaddr_wss is not None or args.boxname is not None:
            raise ValueError(
                "When --wave_csv is specified, None of --data_dir, --ipaddr_wss and --boxname should not be used."
            )
        if not args.wave_csv.exists():
            raise FileNotFoundError(f"Wave stability CSV file not found: {args.wave_csv}")
        return args.wave_csv, tempctrl_csv_path(str(args.wave_csv))

    if args.data_dir is None:
        raise ValueError("Data directory must be specified with --data_dir option when --wave_csv is not provided.")

    if args.ipaddr_wss is not None and args.boxname is None:
        staging_no = int(args.ipaddr_wss.split(".")[-1])
        # possible filename format
        # 1. only ipaddr is specified in data taking
        # 2. boxname is also specified with ipaddr in data taking
        # 3. only ipaddr is specified in data taking (only 4th octet used for file name) This is old file name !
        patterns = [
            f"wavestability_{args.ipaddr_wss.strip().replace('.', '_')}_*.csv",  # 1
            f"wavestability_*_{args.ipaddr_wss.strip().replace('.', '_')}_*.csv",  # 2
            f"wavestability_{staging_no:03}_*.csv",  # 3
        ]
        files = [f for pattern in patterns for f in glob.glob(os.path.join(args.data_dir, "wavestability", pattern))]
    elif args.ipaddr_wss is None and args.boxname is not None:
        files = glob.glob(os.path.join(args.data_dir, "wavestability", f"wavestability_{args.boxname}_*.csv"))
    else:
        raise ValueError("Either --ipaddr or --boxname must be specified when --wave_csv_file is not provided.")

    wave_csv_file_str = max(
        filter(lambda f: validate_filename(f), files),
        key=lambda f: os.path.basename(f).split("_")[-1].split(".")[0],
    )

    return Path(wave_csv_file_str), tempctrl_csv_path(wave_csv_file_str)


def get_tempctrlgraph(
    csv_path: Path, time_window: Tuple[Optional[float], Optional[float]], boxtype: Quel1BoxType
) -> TempCtrlGraph:
    for tcgr in TempCtrlGraph.__subclasses__():
        if tcgr.is_corresponding_boxtype(boxtype):
            return tcgr(csv_path, time_window, boxtype)
    else:
        raise ValueError(f"{args.boxtype} is not supported yet")


def cli_args() -> Namespace:
    logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = ArgumentParser(description="a phase fluctuation measurement tool")
    parser.add_argument(
        "--boxtype",
        type=parse_boxtype,
        required=True,
        help=f"a type of the target box: either of "
        f"{', '.join([t for t in QUEL1_BOXTYPE_ALIAS if not t.startswith('x_')])}",
    )
    parser.add_argument(
        "--ipaddr_wss",
        type=str,
        default=None,
        help="ipaddr_wss specified when stability data was taken",
    )
    parser.add_argument(
        "--boxname",
        type=str,
        help="box name specified when stability data was taken",
    )
    parser.add_argument(
        "--data_dir",
        type=Path,
        default=None,
        help="directory for stability data",
    )
    parser.add_argument(
        "--wave_csv",
        type=Path,
        default=None,
        help="csv file path for wave stability data",
    )
    parser.add_argument(
        "--start_time",
        type=float,
        default=None,
        help="start time of the plots",
    )
    parser.add_argument(
        "--end_time",
        type=float,
        default=None,
        help="end time of the plots",
    )
    parser.add_argument(
        "--freeze",
        action="store_true",
        default=False,
        help="specify for not updating",
    )
    return parser.parse_args()


def cli_body(args: Namespace):
    mpl.use("Gtk3Agg")

    wave_csv_path, temp_csv_path = get_csv_path(args)

    graphs: List[AbstractGraph] = [WaveGraph(wave_csv_path, (args.start_time, args.end_time))]
    if temp_csv_path is not None:
        graphs.append(get_tempctrlgraph(temp_csv_path, (args.start_time, args.end_time), args.boxtype))

    animations = []  # FuncAnimation object must be kept until plt.show()
    if not args.freeze:
        # if args.freeze is False, all graphs will be continuously updated
        # as long as DAQ is on going.
        for graph in graphs:
            animations.append(
                FuncAnimation(  # noqa: F841
                    graph.fig,
                    graph.update_graph,
                    interval=UPDATE_INTERVAL,
                    cache_frame_data=False,
                )
            )
    else:
        # if args.freeze is True,
        # statistical summary about waveform stabilities is displayed.
        for graph in graphs:
            if isinstance(graph, WaveGraph):
                graph.print_wave_stability()

    plt.show()


def cli_main():
    try:
        args = cli_args()
        cli_body(args)
        sys.exit(0)
    except Exception as e:
        logger.error(e)
        sys.exit(1)


if __name__ == "__main__":
    args = cli_args()
    cli_body(args)



import json
import logging
import os
import subprocess
import tempfile
from pathlib import Path
from typing import Optional

import typer
import yaml
from rich.console import Console
from typing_extensions import Annotated

from quel_ic_config_utils import configuration

logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
logger = logging.getLogger()

err_console = Console(stderr=True)

app = typer.Typer(add_completion=False, rich_markup_mode=None)


def _edit_or_abort(original_content: str, suffix: str = ".tmp") -> Optional[str]:
    """
    Opens the user's default editor to modify content.

    Args:
        original_content (str): The initial content to be edited.
        suffix (str): The filename suffix for the temporary file.

    Returns:
        The modified content as a string, or None if no changes were made.
    """
    editor = os.environ.get("EDITOR", "vim")

    with tempfile.NamedTemporaryFile(mode="w+", suffix=suffix, delete=False) as tmp_file:
        tmp_file.write(original_content)
        tmp_file_path = tmp_file.name

    try:
        subprocess.run([editor, tmp_file_path], check=True)

        with open(tmp_file_path, "r") as edited_file:
            edited_content = edited_file.read()

        if edited_content.strip() != original_content.strip():
            return edited_content
        else:
            return None

    except FileNotFoundError as e:
        raise FileNotFoundError(f"Error: Editor '{editor}' not found.") from e
    finally:
        if os.path.exists(tmp_file_path):
            os.remove(tmp_file_path)


def _reconnect_to_box(sysconf_path: Optional[Path], box_name):
    if sysconf_path:
        with open(sysconf_path) as fp:
            conf_object = yaml.safe_load(fp)
        conf = configuration.SystemConfiguration.model_validate(conf_object)
    else:
        conf = configuration.load_default_configuration()

    if conf is None:
        err_console.print("ERROR: Could not find the configuration file.")
        raise typer.Exit(1)

    boxes = list(configuration.get_boxes(b for b in conf.boxes if b.name == box_name))
    if len(boxes) == 0:
        err_console.print(f"ERROR: box '{box_name}' is not found.")
        raise typer.Exit(1)
    box = boxes[0]
    configuration.reconnect_and_get_link_status(box)
    return box


@app.command()
def show(
    box_name: Annotated[str, typer.Argument(help="Name of the Box")],
    sysconf_path: Annotated[
        Optional[Path],
        typer.Option("--sysconf", exists=True, file_okay=True, dir_okay=False, help="System configuration file (yaml)"),
    ] = None,
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            "-l",
            help="Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)",
            case_sensitive=False,
        ),
    ] = "WARNING",
):
    logger.setLevel(getattr(logging, log_level.upper()))

    box = _reconnect_to_box(sysconf_path, box_name)

    config_obj = json.loads(box.dump_box_to_jsonstr())
    formatted_json_string = json.dumps(config_obj, indent=2)

    print(formatted_json_string)


@app.command()
def edit(
    box_name: Annotated[str, typer.Argument(help="Name of the Box")],
    sysconf_path: Annotated[
        Optional[Path],
        typer.Option("--sysconf", exists=True, file_okay=True, dir_okay=False, help="System configuration file (yaml)"),
    ] = None,
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            "-l",
            help="Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)",
            case_sensitive=False,
        ),
    ] = "WARNING",
):
    logger.setLevel(getattr(logging, log_level.upper()))

    box = _reconnect_to_box(sysconf_path, box_name)

    config_obj = json.loads(box.dump_box_to_jsonstr())
    formatted_json_string = json.dumps(config_obj, indent=2)

    del box  # to release lock during editor

    if (modified_str := _edit_or_abort(formatted_json_string)) is None:
        print("Aborted.")
    else:
        try:
            box = _reconnect_to_box(sysconf_path, box_name)
            config_obj = json.loads(box.dump_box_to_jsonstr())
            if formatted_json_string != json.dumps(config_obj, indent=2):
                print("Configuration was changed from another location.")
                print("Aborted.")
                raise typer.Exit(1)

            # to validate json format
            json.loads(modified_str)
        except json.JSONDecodeError as e:
            print(f"JSON decoding error: {e}")
            print("Aborted.")
            raise typer.Exit(1)

        box.config_box_from_jsonstr(modified_str)
        print("Configuration has been updated successfully.")


@app.command()
def overwrite(
    box_name: Annotated[str, typer.Argument(help="Name of the Box")],
    file: Annotated[typer.FileText, typer.Argument(help="Path to JSON file ('-' for stdin)")],
    sysconf_path: Annotated[
        Optional[Path],
        typer.Option("--sysconf", exists=True, file_okay=True, dir_okay=False, help="System configuration file (yaml)"),
    ] = None,
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            "-l",
            help="Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)",
            case_sensitive=False,
        ),
    ] = "WARNING",
):
    logger.setLevel(getattr(logging, log_level.upper()))

    box = _reconnect_to_box(sysconf_path, box_name)

    input_content = file.read()
    try:
        # to validate json format
        json.loads(input_content)
    except json.JSONDecodeError as e:
        err_console.print(f"ERROR: Failed to decode JSON: {e}")
        print("Aborted.")
        raise typer.Exit(1)

    box.config_box_from_jsonstr(input_content)
    print("Configuration has been updated successfully.")


def main():
    app()


if __name__ == "__main__":
    main()



from pybind11.setup_helpers import Pybind11Extension
from setuptools import setup
from glob import glob

ext_modules = [
    Pybind11Extension(
        "adi_ad9082_v170",
        sources=glob("adi_ad9082_v170/*.cpp") + glob("adi_ad9082_v170/ad9082_api/ad9082/src/*.c") + glob("adi_ad9082_v170/ad9082_api/adi_utils/src/*.c"),
        include_dirs=["adi_ad9082_v170/ad9082_api/ad9082/inc", "adi_ad9082_v170/ad9082_api/adi_inc", "adi_ad9082_v170/ad9082_api/adi_utils/inc"],
    )
]

setup(
    ext_modules=ext_modules,
)



import argparse
import logging
from ipaddress import ip_address
from pathlib import Path
from typing import Any

from quel_ic_config import Quel1BoxType
from quel_ic_config_utils import BoxPool, BoxSettingType, VportSettingType, complete_ipaddrs, plot_iqs, single_schedule

logger = logging.getLogger()


def test_loopback(cp: str, pgs: set[str], boxpool: BoxPool, power_thr: dict[str, float]):
    # check_background_noise(cp, power_thr)
    _, iq, chunks = single_schedule(cp, pgs, boxpool, power_thr[cp])
    if len(chunks) != len(pgs):
        logger.error(
            f"the number of pulses captured by the runit '{cp}` is "
            f"expected to be {len(pgs)} but is actually {len(chunks)}, something wrong"
        )
    return iq, chunks


def main():
    logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    for lgrname, lgr in logging.root.manager.loggerDict.items():
        if lgrname in {"root"}:
            pass
        elif lgrname.startswith("quel_ic_config_utils.simple_multibox_framework"):
            pass
        else:
            if isinstance(lgr, logging.Logger):
                lgr.setLevel(logging.WARNING)

    parser = argparse.ArgumentParser(
        "observing the output of port 1, 2, 3, 6, 7, 8, and 9 via internal loop-back paths"
    )
    parser.add_argument(
        "--ipaddr_wss",
        type=ip_address,
        required=True,
        help="IP address of the wave generation/capture subsystem of the target box",
    )
    parser.add_argument(
        "--ipaddr_sss",
        type=ip_address,
        default=0,
        help="IP address of the synchronization subsystem of the target box",
    )
    parser.add_argument(
        "--ipaddr_css",
        type=ip_address,
        default=0,
        help="IP address of the configuration subsystem of the target box",
    )
    parser.add_argument(
        "--png",
        action="store_true",
        help="write graphs into a png file",
    )
    parser.add_argument(
        "--headless",
        action="store_true",
        help="don't show graphs on screen",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)

    CLOCKMASTER_SETTINGS: dict[str, Any] = {}

    BOX_SETTINGS: dict[str, BoxSettingType] = {
        "BOX0": {
            "ipaddr_wss": str(args.ipaddr_wss),
            "ipaddr_sss": str(args.ipaddr_sss),
            "ipaddr_css": str(args.ipaddr_css),
            "boxtype": Quel1BoxType.QuEL1SE_RIKEN8,
            "ignore_crc_error_of_mxfe": {0, 1},
        },
    }

    CAP_VPORT_SETTINGS: dict[str, dict[str, VportSettingType]] = {
        "READ": {
            "create": {
                "boxname": "BOX0",
                "port": 0,  # (0, "r")
                "runits": {0},
            },
            "config": {
                "lo_freq": 8e9,
                "cnco_freq": 1e9,
                "fnco_freq": 0.0,
                "rfswitch": "loop",
            },
            "simple_parameters": {
                0: {
                    "num_delay_sample": 0,
                    "num_integration_section": 1,
                    "num_capture_samples": [3072],
                    "num_blank_samples": [4],
                },
            },
        },
        "MON0": {
            "create": {
                "boxname": "BOX0",
                "port": 4,  # (0, "m")
                "runits": {0},
            },
            "config": {
                "lo_freq": 5e9,
                "cnco_freq": 1e9,
                "fnco_freq": 0.0,
                "rfswitch": "loop",
            },
            "simple_parameters": {
                0: {
                    "num_delay_sample": 0,
                    "num_integration_section": 1,
                    "num_capture_samples": [3072],
                    "num_blank_samples": [4],
                },
            },
        },
        "MON1": {
            "create": {
                "boxname": "BOX0",
                "port": 10,  # (1, "m")
                "runits": {0},
            },
            "config": {
                "lo_freq": 5e9,
                "cnco_freq": 1e9,
                "fnco_freq": 0.0,
                "rfswitch": "loop",
            },
            "simple_parameters": {
                0: {
                    "num_delay_sample": 0,
                    "num_integration_section": 1,
                    "num_capture_samples": [3072],
                    "num_blank_samples": [4],
                },
            },
        },
    }

    GEN_VPORT_SETTINGS: dict[str, dict[str, VportSettingType]] = {
        "GEN00": {
            "create": {
                "boxname": "BOX0",
                "port": 1,
                "channel": 0,
            },
            "config": {
                "lo_freq": 8e9,
                "cnco_freq": 1e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
                "sideband": "L",
                "vatt": 0xC00,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 64,
                "num_repeats": (1, 1),
                "num_wait_samples": (0, 0),
            },
        },
        "GEN01": {
            "create": {
                "boxname": "BOX0",
                "port": (1, 1),
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 64,
                "num_repeats": (1, 1),
                "num_wait_samples": (0, 0),
            },
        },
        "GEN02": {
            "create": {
                "boxname": "BOX0",
                "port": 2,
                "channel": 0,
            },
            "config": {
                "lo_freq": 8e9,
                "cnco_freq": 2e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
                "sideband": "L",
                "vatt": 0xC00,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 128,
                "num_repeats": (1, 1),
                "num_wait_samples": (512, 0),
            },
        },
        "GEN03": {
            "create": {
                "boxname": "BOX0",
                "port": 3,
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 64,
                "num_repeats": (1, 1),
                "num_wait_samples": (1024, 0),
            },
        },
        "GEN06": {
            "create": {
                "boxname": "BOX0",
                "port": 6,
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 64,
                "num_repeats": (1, 1),
                "num_wait_samples": (0, 0),
            },
        },
        "GEN07": {
            "create": {
                "boxname": "BOX0",
                "port": 7,
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 128,
                "num_repeats": (1, 1),
                "num_wait_samples": (512, 0),
            },
        },
        "GEN08": {
            "create": {
                "boxname": "BOX0",
                "port": 8,
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 64,
                "num_repeats": (1, 1),
                "num_wait_samples": (1024, 0),
            },
        },
        "GEN09": {
            "create": {
                "boxname": "BOX0",
                "port": 9,
                "channel": 0,
            },
            "config": {
                "cnco_freq": 4e9,
                "fnco_freq": 0.0,
                "fullscale_current": 40000,
            },
            "cw_parameter": {
                "amplitude": 32767.0,
                "num_wave_sample": 128,
                "num_repeats": (1, 1),
                "num_wait_samples": (1536, 0),
            },
        },
    }

    boxpool0 = BoxPool(CLOCKMASTER_SETTINGS, BOX_SETTINGS, CAP_VPORT_SETTINGS, GEN_VPORT_SETTINGS)
    boxpool0.initialize(allow_resync=False)

    bgnoise = boxpool0.check_background_noise({"READ", "MON0", "MON1"})
    bgnoise_thr = {
        "READ": 2000.0,
        "MON0": 200.0,
        "MON1": 200.0,
    }
    for runit_name in bgnoise:
        if bgnoise[runit_name] > bgnoise_thr[runit_name] * 0.75:
            logger.warning(
                f"the runit {runit_name} is too noisy for the given power threshold of "
                f"pulse detection (= {bgnoise_thr[runit_name]}), you may see spurious pulses in the results"
            )

    boxpool0.measure_timediff("MON1")

    iq0, chunk0 = test_loopback("READ", {f"GEN{idx:02d}" for idx in (0,)}, boxpool0, bgnoise_thr)
    iq1, chunk1 = test_loopback("MON0", {f"GEN{idx:02d}" for idx in (1, 2, 3)}, boxpool0, bgnoise_thr)
    iq2, chunk2 = test_loopback("MON1", {f"GEN{idx:02d}" for idx in (6, 7, 8, 9)}, boxpool0, bgnoise_thr)

    plot_iqs(
        {
            "read-loop: port-#01": iq0,
            "monitor0-loop: port-#01, port-#02, port-#03": iq1,
            "monitor1-loop: port-#06, port-#07, port-#08, port-#09": iq2,
        },
        same_range=False,
        show_graph=not args.headless,
        output_filename=Path(str(args.ipaddr_wss) + ".png") if args.png else None,
    )

    del boxpool0


if __name__ == "__main__":
    main()



import logging
from abc import ABCMeta
from typing import ClassVar, Final, Optional, Type, Union

from pydantic import BaseModel, Field, field_validator, model_validator

from quel_ic_config import (
    Quel1BoxType,
    Quel1NormalThermistor,
    Quel1PathSelectorThermistor,
    Quel1seFujitsu11TypeAConfigSubsystem,
    Quel1seFujitsu11TypeBConfigSubsystem,
    Quel1seRiken8ConfigSubsystem,
    Quel1Thermistor,
)

logger = logging.getLogger(__name__)

QUEL1_THERMISTORS: dict[int, Quel1Thermistor] = {
    0: Quel1NormalThermistor("adrf6780_0_0"),
    1: Quel1NormalThermistor("adrf6780_0_1"),
    2: Quel1NormalThermistor("adrf6780_0_2"),
    3: Quel1NormalThermistor("adrf6780_0_3"),
    4: Quel1NormalThermistor("adrf6780_1_0"),
    5: Quel1NormalThermistor("adrf6780_1_1"),
    6: Quel1NormalThermistor("adrf6780_1_2"),
    7: Quel1NormalThermistor("adrf6780_1_3"),
    8: Quel1NormalThermistor("ad9082_0"),
    9: Quel1NormalThermistor("ad9082_1"),
    10: Quel1NormalThermistor("lmx2594_0_0"),
    11: Quel1NormalThermistor("lmx2594_0_1"),
    12: Quel1NormalThermistor("lmx2594_0_2"),
    13: Quel1NormalThermistor("lmx2594_0_3"),
    14: Quel1NormalThermistor("lmx2594_1_0"),
    15: Quel1NormalThermistor("lmx2594_1_1"),
    16: Quel1NormalThermistor("lmx2594_1_2"),
    17: Quel1NormalThermistor("lmx2594_1_3"),
    18: Quel1NormalThermistor("lmx2594_12g_0"),
    19: Quel1NormalThermistor("lmx2594_12g_1"),
    20: Quel1NormalThermistor("adclk950"),
    21: Quel1NormalThermistor("adclk925"),
    22: Quel1NormalThermistor("rx_0_0"),
    23: Quel1NormalThermistor("rx_0_1"),
    24: Quel1NormalThermistor("rx_1_0"),
    25: Quel1NormalThermistor("rx_1_1"),
    26: Quel1PathSelectorThermistor("ps_0"),
    27: Quel1PathSelectorThermistor("ps_1"),
}

QUEL1_ACTUATORS: dict[str, tuple[str, int]] = {
    "adrf6780_0_0": ("peltier", 0),
    "adrf6780_0_1": ("peltier", 1),
    "adrf6780_0_2": ("peltier", 2),
    "adrf6780_0_3": ("peltier", 3),
    "adrf6780_1_0": ("peltier", 4),
    "adrf6780_1_1": ("peltier", 5),
    "adrf6780_1_2": ("peltier", 6),
    "adrf6780_1_3": ("peltier", 7),
    "ad9082_0": ("peltier", 8),
    "ad9082_1": ("peltier", 9),
    "lmx2594_0_0": ("peltier", 10),
    "lmx2594_0_1": ("peltier", 11),
    "lmx2594_0_2": ("peltier", 12),
    "lmx2594_0_3": ("peltier", 13),
    "lmx2594_1_0": ("peltier", 14),
    "lmx2594_1_1": ("peltier", 15),
    "lmx2594_1_2": ("peltier", 16),
    "lmx2594_1_3": ("peltier", 17),
    "lmx2594_12g_0": ("peltier", 18),
    "lmx2594_12g_1": ("peltier", 19),
    "adclk950": ("peltier", 20),
    "adclk925": ("peltier", 21),
    "rx_0_0": ("peltier", 22),  # Notes: inverted peltier
    "rx_0_1": ("peltier", 23),  # Notes: inverted peltier
    "rx_1_0": ("peltier", 24),  # Notes: inverted peltier
    "rx_1_1": ("peltier", 25),  # Notes: inverted peltier
    "ps_0": ("heater", 26),
    "ps_1": ("heater", 27),
}

_MXFE_THERMISTORS: Final[set[str]] = {"mxfe0_max", "mxfe0_min", "mxfe1_max", "mxfe1_min"}


class QuelDataModel(BaseModel, extra="forbid", metaclass=ABCMeta):
    time_from_start: float = Field(ge=0.0)


class WaveStatistics(QuelDataModel):
    output_port: str = Field()
    input_port: str = Field()
    power_mean: float = Field(ge=0, le=32767)
    power_std: float = Field(ge=0, le=32767)
    angle_mean: float = Field(
        ge=-180.0, lt=188.2
    )  # (pi + (pi-3.0)) * 180/pi = 188.11 see WaveStabilityMeasurement.phase_stat in wave_stability_check.py
    angle_std: float = Field(ge=0, le=180.0)  # max standard deviation in the range [-pi,pi] is (pi-(-pi))/2 = pi
    angle_deltamax: float = Field(ge=0, le=360.0)


class TempCtrlStatus(QuelDataModel):
    location_name: str = Field()
    temperature: float = Field(ge=0, lt=120)
    # Notes: for thermistors that doesn't have its proximal thermal actuator, actuator_type should be None.
    actuator_type: Optional[str] = Field(default=None)
    # Notes: for locations of which actuator_type is None, actuator_val should be None, either.
    actuator_val: Optional[float] = Field(default=None)

    _LOCATION_NAMES: ClassVar[set[str]] = set()
    _VALID_ACTUATOR_TYPES: ClassVar[set[Union[str, None]]] = set()
    _CORRESPONDING_BOXTYPES: ClassVar[set[Quel1BoxType]] = set()

    @field_validator("location_name")
    def validate_location(cls, val):
        if val not in cls._LOCATION_NAMES:
            raise ValueError(f"no such thermistor or actuator location {val}")
        return val

    @field_validator("actuator_type")
    def validate_actuator_type(cls, val):
        if val not in cls._VALID_ACTUATOR_TYPES:
            raise ValueError(
                f"actuator type must be either of {', '.join([str(at) for at in cls._VALID_ACTUATOR_TYPES])}"
            )
        return val

    @classmethod
    def is_corresponding_boxtype(cls, boxtype: Quel1BoxType) -> bool:
        return boxtype in cls._CORRESPONDING_BOXTYPES


class Quel1TempCtrlStatus(TempCtrlStatus):
    _LOCATION_NAMES: ClassVar[set[str]] = set(QUEL1_ACTUATORS.keys())
    _VALID_ACTUATOR_TYPES: ClassVar[set[Union[str, None]]] = {"heater", "peltier"}

    # Notes: be sure that QuBE doesn't have CMOD-35T with uart_monitor firmware.
    _CORRESPONDING_BOXTYPES: ClassVar[set[Quel1BoxType]] = {
        Quel1BoxType.QuEL1_TypeA,
        Quel1BoxType.QuEL1_TypeB,
        Quel1BoxType.QuEL1_NEC,
        Quel1BoxType.QuEL1_NTT,
    }

    @model_validator(mode="after")
    @classmethod
    def validate_actuator_val(cls, model):
        if model.actuator_type == "peltier":
            if model.actuator_val is None or not (1 <= model.actuator_val <= 200):
                raise ValueError("actuator value for peltiers must be within 1-200")
        elif model.actuator_type == "heater":
            if model.actuator_val is None or not (1 <= model.actuator_val <= 300):
                raise ValueError("actuator value for heaters must be within 1-300")
        else:
            # Notes: Notes: Never happens
            raise AssertionError("never happens")

        return model


class Quel1seTempCtrlStatus(TempCtrlStatus):
    _VALID_ACTUATOR_TYPES: ClassVar[set[Union[str, None]]] = {"fan", "heater", None}

    @field_validator("actuator_val")
    def validate_actuator_val(cls, val):
        if val and not (0.0 <= val <= 1.0):
            raise ValueError("actuator value must be in the range btwn 0.0 and 1.0")
        return val


class Quel1seRiken8TempCtrlStatus(Quel1seTempCtrlStatus):
    _LOCATION_NAMES: ClassVar[set[str]] = {
        th.name for th in Quel1seRiken8ConfigSubsystem._THERMISTORS.values()
    } | _MXFE_THERMISTORS
    _CORRESPONDING_BOXTYPES: ClassVar[set[Quel1BoxType]] = {Quel1BoxType.QuEL1SE_RIKEN8, Quel1BoxType.QuEL1SE_RIKEN8DBG}


class Quel1seFujitsu11TypeATempCtrlStatus(Quel1seTempCtrlStatus):
    _LOCATION_NAMES: ClassVar[set[str]] = {
        thermistor.name for thermistor in Quel1seFujitsu11TypeAConfigSubsystem._THERMISTORS.values()
    } | _MXFE_THERMISTORS
    _CORRESPONDING_BOXTYPES: ClassVar[set[Quel1BoxType]] = {
        Quel1BoxType.QuEL1SE_FUJITSU11_TypeA,
        Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeA,
    }


class Quel1seFujitsu11TypeBTempCtrlStatus(Quel1seTempCtrlStatus):
    _LOCATION_NAMES: ClassVar[set[str]] = {
        thermistor.name for thermistor in Quel1seFujitsu11TypeBConfigSubsystem._THERMISTORS.values()
    } | _MXFE_THERMISTORS
    _CORRESPONDING_BOXTYPES: ClassVar[set[Quel1BoxType]] = {
        Quel1BoxType.QuEL1SE_FUJITSU11_TypeB,
        Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeB,
    }


def _get_all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in _get_all_subclasses(c)])


def get_tempctrlstatus_from_boxtype(boxtype: Quel1BoxType) -> Type[TempCtrlStatus]:
    for tcscls in _get_all_subclasses(TempCtrlStatus):
        if tcscls.is_corresponding_boxtype(boxtype):
            return tcscls
    raise ValueError(f"unsupported boxtype: {Quel1BoxType.tostr(boxtype)}")



import logging
import os
import time
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Final, List, Set, Tuple, Union

import numpy as np
from e7awghal import CapParam, CapSection, E7awgCaptureDataError

from quel_ic_config.ad9082 import LinkStatus
from quel_ic_config.e7resource_mapper import AbstractQuel1E7ResourceMapper
from quel_ic_config.quel1_any_config_subsystem import Quel1AnyConfigSubsystem
from quel_ic_config.quel1_config_subsystem_tempctrl import Quel1seTempctrlState
from quel_ic_config.quel1_wave_subsystem import Quel1WaveSubsystem

logger = logging.getLogger(__name__)


class LinkupStatus(Enum):
    LINKUP_FAILURE = 1
    CRC_ERROR = 2
    RX_TIMEOUT = 3
    RX_FAILURE = 4
    EMPTY_RX_DATA = 5
    BROKEN_RX_DATA = 6
    CHIKUCHIKU = 7
    GOOD_CAPTURE = 8


@dataclass(frozen=True)
class LinkupStatistic:
    link_status: LinkStatus
    error_status: int
    fddc_idx: int
    timeout: bool
    failure: bool
    samples: int
    valid_capture: bool
    max_noise_peak: float

    def categorize(self, noise_threshold: float) -> LinkupStatus:
        if self.link_status != LinkStatus.LINK_STATUS_LOCKED:
            return LinkupStatus.LINKUP_FAILURE
        elif self.error_status != 0x01:
            return LinkupStatus.CRC_ERROR
        elif self.failure:
            return LinkupStatus.RX_FAILURE
        elif self.timeout:
            return LinkupStatus.RX_TIMEOUT
        elif not self.valid_capture:
            if self.samples == 0:
                return LinkupStatus.EMPTY_RX_DATA
            else:
                return LinkupStatus.BROKEN_RX_DATA
        elif self.max_noise_peak >= noise_threshold:
            return LinkupStatus.CHIKUCHIKU
        else:
            return LinkupStatus.GOOD_CAPTURE

    def is_success(self, noise_threshold: float) -> bool:
        return self.categorize(noise_threshold) == LinkupStatus.GOOD_CAPTURE


class LinkupFpgaMxfe:
    _LINKUP_MAX_RETRY: Final[int] = 10
    _SLEEP_BTWN_LINKUP_TRIALS: Final[float] = 0.25
    _DEFAULT_BACKGROUND_NOISE_THRESHOLD_RELINKUP: Final[float] = 256.0
    _DEFAULT_BACKGROUND_NOISE_THRESHOLD_AT_RECONNECT: Final[float] = 4096.0
    _STAT_HISTORY_MAX_LEN: Final[int] = 1000

    def __init__(self, css: Quel1AnyConfigSubsystem, wss: Quel1WaveSubsystem, rmap: AbstractQuel1E7ResourceMapper):
        self._css = css
        self._wss = wss
        self._rmap: AbstractQuel1E7ResourceMapper = rmap
        self._statistics: Dict[int, List[LinkupStatistic]] = {}
        self._target_capmods: Dict[int, Set[int]] = {}

    def _validate_mxfe(self, mxfe_idx: int) -> bool:
        return mxfe_idx in self._css.get_all_mxfes()

    def _get_fddcs_of_mxfe(self, mxfe_idx: int) -> Set[int]:
        fddcs: Set[int] = set()
        for g in self._css.get_all_groups():
            for rl in self._css.get_all_rlines_of_group(g):
                for rch in range(self._css.get_num_rchannels_of_rline(g, rl)):
                    m, d = self._css.get_fddc_idx(g, rl, rch)
                    if m == mxfe_idx:
                        fddcs.add(d)
        return fddcs

    def _add_linkup_statistics(
        self,
        mxfe_idx: int,
        fddc_idx: int,
        timeout: bool,
        failure: bool,
        samples: int,
        valid_capture: bool,
        max_noise_peak: float,
    ):
        link_status, crc_error_status = self._css.get_link_status(mxfe_idx)
        if mxfe_idx not in self._statistics:
            self._statistics[mxfe_idx] = []

        self._statistics[mxfe_idx].append(
            LinkupStatistic(
                link_status=link_status,
                error_status=crc_error_status,
                fddc_idx=fddc_idx,
                timeout=timeout,
                failure=failure,
                samples=samples,
                valid_capture=valid_capture,
                max_noise_peak=max_noise_peak,
            )
        )

        if len(self._statistics[mxfe_idx]) > self._STAT_HISTORY_MAX_LEN:
            self._statistics[mxfe_idx] = self._statistics[mxfe_idx][-self._STAT_HISTORY_MAX_LEN :]  # noqa: E203

    def clear_statistics(self, mxfe_idx: Union[None, int] = None):
        if mxfe_idx is None:
            self._statistics = {}
        else:
            self._validate_mxfe(mxfe_idx)
            if mxfe_idx not in self._statistics:
                logger.warning(f"no data exist for mxfe-#{mxfe_idx}, do nothing actually")
            self._statistics[mxfe_idx] = []

    @property
    def linkup_statistics(self):
        return self._statistics

    def init_wss_resources(self, mxfe_idx: int) -> None:
        self._validate_mxfe(mxfe_idx)
        self._wss.initialize_awgunits(self._rmap.get_awgs_of_mxfe(mxfe_idx))
        capunits: Set[Tuple[int, int]] = set()
        for fddc_idx in self._get_fddcs_of_mxfe(mxfe_idx):
            capunits.add((self._rmap.get_capmod_from_fddc(mxfe_idx, fddc_idx), 0))
        self._wss.initialize_capunits(capunits)

    def linkup_and_check(
        self,
        mxfe_idx: int,
        param: Dict[str, Any],
        *,
        hard_reset: bool = False,
        soft_reset: bool = True,
        use_204b: bool = False,
        use_bg_cal: bool = True,
        ignore_crc_error: bool = False,
        ignore_extraordinary_converter_select: bool = False,
        background_noise_threshold: Union[float, None] = None,
        restart_tempctrl: bool = False,
        save_dirpath: Union[Path, None] = None,
    ) -> bool:
        self._validate_mxfe(mxfe_idx)
        if background_noise_threshold is None:
            background_noise_threshold = self._DEFAULT_BACKGROUND_NOISE_THRESHOLD_RELINKUP

        judge_system: bool = False
        for i in range(self._LINKUP_MAX_RETRY):
            if i != 0:
                logger.debug(f"waiting {self._SLEEP_BTWN_LINKUP_TRIALS} seconds before retrying linkup")
                time.sleep(self._SLEEP_BTWN_LINKUP_TRIALS)

            if not self._css.configure_mxfe(
                mxfe_idx,
                param,
                hard_reset=hard_reset,
                soft_reset=soft_reset,
                use_204b=use_204b,
                use_bg_cal=use_bg_cal,
                ignore_crc_error=ignore_crc_error,
            ):
                self._add_linkup_statistics(
                    mxfe_idx,
                    fddc_idx=-1,
                    timeout=False,
                    failure=False,
                    samples=-1,
                    valid_capture=False,
                    max_noise_peak=-1,
                )
                continue

            self.init_wss_resources(mxfe_idx)

            # Notes: it is fine to check all the available adcs of the target mxfe.
            if self.check_all_fddcs_of_mxfe_at_relinkup(mxfe_idx, background_noise_threshold, save_dirpath):
                logger.info(f"successful system-level link-up of {self._css.ipaddr_css}:mxfe-#{mxfe_idx}")
                judge_system = True
                break
            else:
                # Notes: info is enough
                logger.info(f"failed system-level link-up of {self._css.ipaddr_css}:mxfe-#{mxfe_idx}")

        if judge_system:
            if self._css.tempctrl_auto_start_at_linkup:
                self.start_temperature_control(restart_tempctrl)

        return judge_system

    def check_all_fddcs_of_mxfe_at_relinkup(
        self,
        mxfe_idx: int,
        background_noise_threshold: Union[float, None] = None,
        save_dirpath: Union[Path, None] = None,
    ) -> bool:
        judge_fddcs: bool = True
        for fddc_idx in self._get_fddcs_of_mxfe(mxfe_idx):
            judge_fddcs &= self.check_fddc(
                mxfe_idx,
                fddc_idx,
                background_noise_threshold or self._DEFAULT_BACKGROUND_NOISE_THRESHOLD_RELINKUP,
                save_dirpath,
            )
        return judge_fddcs

    def check_all_fddcs_of_mxfe_at_reconnect(
        self,
        mxfe_idx: int,
        background_noise_threshold: Union[float, None] = None,
        save_dirpath: Union[Path, None] = None,
    ) -> bool:
        judge_fddcs: bool = True
        for fddc_idx in self._get_fddcs_of_mxfe(mxfe_idx):
            judge_fddcs &= self.check_fddc(
                mxfe_idx,
                fddc_idx,
                background_noise_threshold or self._DEFAULT_BACKGROUND_NOISE_THRESHOLD_AT_RECONNECT,
                save_dirpath,
            )
        return judge_fddcs

    def check_fddc(
        self,
        mxfe_idx: int,
        fddc_idx: int,
        background_noise_threshold: float,
        save_dirpath: Union[Path, None] = None,
    ) -> bool:
        capmod = self._rmap.get_capmod_from_fddc(mxfe_idx, fddc_idx)
        capunit = (capmod, 0)
        param = CapParam(num_repeat=1)
        param.sections.append(CapSection(name="s0", num_capture_word=16384, num_blank_word=1))
        self._wss.config_capunit(capunit, param)
        task = self._wss.start_capunits_now({capunit})
        try:
            reader = task.result()[capunit]
        except TimeoutError as e:
            logger.warning(e)
            self._add_linkup_statistics(
                mxfe_idx,
                fddc_idx=fddc_idx,
                timeout=True,
                failure=False,
                samples=-1,
                valid_capture=False,
                max_noise_peak=-1,
            )
            return False
        except E7awgCaptureDataError as e:
            logger.warning(e)
            self._add_linkup_statistics(
                mxfe_idx,
                fddc_idx=fddc_idx,
                timeout=False,
                failure=False,
                samples=self._wss._get_capunit(capunit).get_num_captured_sample(),  # TODO: improve the design.
                valid_capture=False,
                max_noise_peak=-1,
            )
            return False

        cap_data = reader.as_wave_dict()["s0"][0]
        if save_dirpath is not None:
            os.makedirs(save_dirpath, exist_ok=True)
            np.save(str(save_dirpath / f"backgroud_{mxfe_idx}_{fddc_idx}_{int(time.time())}.npy"), cap_data)

        max_backgroud_amplitude = max(abs(cap_data))
        if max_backgroud_amplitude < background_noise_threshold:
            logger.info(
                f"successful establishment of capture link of fddc-#{fddc_idx} of mxfe-#{mxfe_idx}, "
                f"max amplitude of the capture data is "
                f"{max_backgroud_amplitude:.1f} (< {background_noise_threshold:.1f})"
            )
            self._add_linkup_statistics(
                mxfe_idx=mxfe_idx,
                fddc_idx=fddc_idx,
                timeout=False,
                failure=False,
                samples=cap_data.shape[0],
                valid_capture=True,
                max_noise_peak=max_backgroud_amplitude,
            )
            return True
        else:
            # need to link up again to make the captured data fine
            logger.warning(
                f"failed establishment of capture link of fddc-#{fddc_idx} of mxfe-#{mxfe_idx}, "
                "max amplitude of the capture data of is "
                f"{max_backgroud_amplitude:.1f} (>= {background_noise_threshold:.1f})"
            )
            self._add_linkup_statistics(
                mxfe_idx=mxfe_idx,
                fddc_idx=fddc_idx,
                timeout=False,
                failure=False,
                samples=cap_data.shape[0],
                valid_capture=True,
                max_noise_peak=max_backgroud_amplitude,
            )
            return False

    def start_temperature_control(self, restart: bool) -> None:
        tc_state = self._css.get_tempctrl_state()
        if tc_state == Quel1seTempctrlState.INIT:
            logger.info(f"the temperature control is initialized ({tc_state.tostr()} -> prerun)")
            self._css.start_tempctrl()
        elif tc_state in {Quel1seTempctrlState.PRERUN, Quel1seTempctrlState.RUN}:
            if restart:
                logger.info(f"the temperature control is initialized ({tc_state.tostr()} -> prerun)")
                self._css.start_tempctrl()
        else:
            logger.warning(f"be aware that the temperature control is in an experimental mode ({tc_state.tostr()})")



import logging
import time
from pathlib import Path
from typing import Any, Collection, Dict, Final, Mapping, Set, Tuple, Union

from quel_ic_config.exstickge_sock_client import (
    AbstractLockKeeper,
    DummyLockKeeper,
    FileLockKeeper,
    LsiKindId,
    _ExstickgeSockClientBase,
)
from quel_ic_config.quel1_config_subsystem_common import (
    Quel1ConfigSubsystemAd5328Mixin,
    Quel1ConfigSubsystemAd6780Mixin,
    Quel1ConfigSubsystemLmx2594Mixin,
    Quel1ConfigSubsystemNoRfswitch,
    Quel1ConfigSubsystemRfswitch,
    Quel1ConfigSubsystemRoot,
    Quel1GenericConfigSubsystemAd9082Mixin,
)
from quel_ic_config.quel1_config_subsystem_tempctrl import Quel1ConfigSubsystemTempctrlMixin
from quel_ic_config.quel_config_common import _DEFAULT_LOCK_DIRECTORY, Quel1BoxType
from quel_ic_config.quel_ic import Ad9082Generic

_DEFAULT_SOCKET_SERVER_DETECTION_TIMEOUT: Final[float] = 15.0  # [s]

logger = logging.getLogger(__name__)


class AbstractExstickgeSockClientQuel1(_ExstickgeSockClientBase):
    _AD9082_IF_0 = LsiKindId.AD9082
    _ADRF6780_IF_0 = LsiKindId.ADRF6780
    _ADRF6780_IF_1 = LsiKindId.ADRF6780 + 1
    _LMX2594_IF_0 = LsiKindId.LMX2594
    _LMX2594_IF_1 = LsiKindId.LMX2594 + 1
    _AD5328_IF_0 = LsiKindId.AD5328
    _GPIO_IF_0 = LsiKindId.GPIO

    _SPIIF_MAPPINGS: Mapping[LsiKindId, Mapping[int, Tuple[int, int]]] = {
        LsiKindId.AD9082: {
            0: (_AD9082_IF_0, 0),
            1: (_AD9082_IF_0, 1),
        },
        LsiKindId.ADRF6780: {
            0: (_ADRF6780_IF_0, 0),
            1: (_ADRF6780_IF_0, 1),
            2: (_ADRF6780_IF_0, 2),
            3: (_ADRF6780_IF_0, 3),
            4: (_ADRF6780_IF_1, 0),
            5: (_ADRF6780_IF_1, 1),
            6: (_ADRF6780_IF_1, 2),
            7: (_ADRF6780_IF_1, 3),
        },
        LsiKindId.LMX2594: {
            0: (_LMX2594_IF_0, 0),
            1: (_LMX2594_IF_0, 1),
            2: (_LMX2594_IF_0, 2),
            3: (_LMX2594_IF_0, 3),
            4: (_LMX2594_IF_0, 4),
            5: (_LMX2594_IF_1, 0),
            6: (_LMX2594_IF_1, 1),
            7: (_LMX2594_IF_1, 2),
            8: (_LMX2594_IF_1, 3),
            9: (_LMX2594_IF_1, 4),
        },
        LsiKindId.AD5328: {
            0: (_AD5328_IF_0, 0),
        },
        LsiKindId.GPIO: {
            0: (_GPIO_IF_0, 0),
        },
    }

    def __init__(
        self,
        target_address,
        target_port=_ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        receiver_limit_by_binding: bool = False,
    ):
        super().__init__(target_address, target_port, timeout, receiver_limit_by_binding)


class ExstickgeSockClientQuel1WithDummyLock(AbstractExstickgeSockClientQuel1):
    def _create_lockkeeper(self) -> AbstractLockKeeper:
        t = DummyLockKeeper(target=self._target)
        t.activate()
        return t


class ExstickgeSockClientQuel1WithFileLock(AbstractExstickgeSockClientQuel1):
    def __init__(
        self,
        target_address,
        target_port=_ExstickgeSockClientBase.DEFAULT_PORT,
        lock_directory: Path = _DEFAULT_LOCK_DIRECTORY,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        receiver_limit_by_binding: bool = False,
    ):
        self._lock_directory: Path = lock_directory
        super().__init__(target_address, target_port, timeout, receiver_limit_by_binding)

    def _create_lockkeeper(self) -> AbstractLockKeeper:
        t = FileLockKeeper(target=self._target, lock_directory=self._lock_directory)
        t.activate()
        return t


class Ad9082Quel1(Ad9082Generic):
    def _reset_pin_ctrl_cb(self, level: int) -> Tuple[bool]:
        logger.warning("hardware reset of AD9082 is not implemented")
        return (False,)


class Quel1ConfigSubsystemAd9082Mixin(Quel1GenericConfigSubsystemAd9082Mixin):
    def _construct_ad9082(self):
        self._ad9082: tuple[Ad9082Generic, ...] = tuple(
            Ad9082Quel1(self._proxy, idx) for idx in range(self._NUM_IC["ad9082"])
        )
        self.allow_dual_modulus_nco = True


class QuelMeeBoardConfigSubsystem(
    Quel1ConfigSubsystemRoot,
    Quel1ConfigSubsystemAd9082Mixin,
    Quel1ConfigSubsystemLmx2594Mixin,
    Quel1ConfigSubsystemAd6780Mixin,
    Quel1ConfigSubsystemAd5328Mixin,
    Quel1ConfigSubsystemTempctrlMixin,
):
    __slots__ = ()

    _DEFAULT_CONFIG_JSONFILE: str = "quel-1.json"

    # TODO: move "gpio" to the appropriate place.
    _NUM_IC: Dict[str, int] = {
        "ad9082": 2,
        "lmx2594": 10,
        "adrf6780": 8,
        "ad5328": 1,
        "gpio": 1,
    }

    _GROUPS: Set[int]

    _MXFE_IDXS: Set[int] = {0, 1}
    _LMX2594_OF_MXFES: Tuple[int, ...] = (8, 9)

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        Quel1ConfigSubsystemRoot.__init__(self, css_addr, boxtype, port, timeout, sender_limit_by_binding)

    def initialize(self) -> None:
        super().initialize()
        self._construct_ad9082()
        self._construct_lmx2594()
        self._construct_adrf6780()
        self._construct_ad5328()

    def _create_exstickge_proxy(
        self, port: int, timeout: float, sender_limit_by_binding: bool
    ) -> _ExstickgeSockClientBase:
        # TODO: consider to accept user defined lock directory or not

        proxy = ExstickgeSockClientQuel1WithFileLock(
            self._css_addr, port, _DEFAULT_LOCK_DIRECTORY, timeout, sender_limit_by_binding
        )
        proxy.initialize()
        # Notes: check the availablity of some end-points of the server after taking the lock.
        #        MEE board has LMX2594[0] definitely.
        # Notes: lock should be acquired at the end of ExstickgeSockClientQuel1WithFileLock.__init__() if available.
        if proxy.has_lock:
            t0 = time.perf_counter()
            while time.perf_counter() < t0 + _DEFAULT_SOCKET_SERVER_DETECTION_TIMEOUT:
                if proxy.read_reg(LsiKindId.LMX2594, 0, 0x0000) is not None:
                    break
                time.sleep(3.0)
            else:
                raise RuntimeError(f"socket server is not available on {self._css_addr}")
        else:
            # Notes: BoxLockError will raise at _create_css_object()
            pass

        return proxy

    def configure_peripherals(
        self,
        param: Dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None:
        if ignore_access_failure_of_adrf6780 is None:
            ignore_access_failure_of_adrf6780 = {}

        if ignore_lock_failure_of_lmx2594 is None:
            ignore_lock_failure_of_lmx2594 = {}

        self.init_ad5328(0, param["ad5328"][0])
        for i in range(0, 8):
            self.init_adrf6780(i, param["adrf6780"][i], ignore_id_mismatch=i in ignore_access_failure_of_adrf6780)
            self.init_lmx2594(i, param["lmx2594"][i], ignore_lock_failure=i in ignore_lock_failure_of_lmx2594)

    def configure_all_mxfe_clocks(
        self, param: Dict[str, Any], *, ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None
    ) -> None:
        if ignore_lock_failure_of_lmx2594 is None:
            ignore_lock_failure_of_lmx2594 = {}

        for group in range(2):
            lmx2594_idx = 8 + group
            self.init_lmx2594(
                lmx2594_idx,
                param["lmx2594"][lmx2594_idx],
                ignore_lock_failure=lmx2594_idx in ignore_lock_failure_of_lmx2594,
            )

    def configure_mxfe(
        self,
        mxfe_idx: int,
        param: Dict[str, Any],
        *,
        hard_reset: bool = False,
        soft_reset: bool = False,
        use_204b: bool = False,
        use_bg_cal: bool = True,
        ignore_crc_error: bool = False,
    ) -> bool:
        self._validate_mxfe(mxfe_idx)

        if hard_reset:
            logger.warning(
                f"QuEL-1 ({self._css_addr}) does not support hardware reset of AD9082-#{mxfe_idx}, "
                "conducts software reset instead."
            )
            hard_reset = False
            soft_reset = True

        self.ad9082[mxfe_idx].configure(
            param["ad9082"][mxfe_idx],
            hard_reset=hard_reset,
            soft_reset=soft_reset,
            use_204b=use_204b,
            use_bg_cal=use_bg_cal,
        )
        health, lglv, diag = self.check_link_status(mxfe_idx, True, ignore_crc_error)
        logger.log(lglv, diag)
        return health

    def reconnect_mxfe(
        self,
        mxfe_idx: int,
        *,
        ignore_crc_error: bool = False,
    ) -> bool:
        self._validate_mxfe(mxfe_idx)
        pll = self.lmx2594[self._LMX2594_OF_MXFES[mxfe_idx]]
        dr = pll.get_divider_ratio()[0]  # Notes: assuming that OUTA is used.
        se = pll.get_sync_enable()
        if dr == 0:
            raise RuntimeError(f"PLL of AD9082-#{mxfe_idx} is not activated")
        ref_clk = pll.get_lo_multiplier() * 100_000_000 // dr
        if se:
            ref_clk *= 4

        self.ad9082[mxfe_idx].reconnect(ref_clk)
        health, lglv, diag = self.check_link_status(mxfe_idx, False, ignore_crc_error)
        logger.log(lglv, diag)
        return health

    def dump_channel(self, group: int, line: int, channel: int) -> Dict[str, Any]:
        """dumping the current configuration of a transmitter channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the target channel belongs to.
        :param channel: a line-local index of the channel.
        :return: the current configuration information of the channel.
        """
        self._validate_channel(group, line, channel)
        return {
            "fnco_freq": self.get_dac_fnco(group, line, channel),
        }

    def dump_line(self, group: int, line: int) -> Dict[str, Any]:
        """dumping the current configuration of a transmitter line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: the current configuration information of the line.
        """
        self._validate_line(group, line)
        r: Dict[str, Any] = {}
        r["channels"] = {
            ch: self.dump_channel(group, line, ch) for ch in range(self.get_num_channels_of_line(group, line))
        }
        r["cnco_freq"] = self.get_dac_cnco(group, line)
        r["fullscale_current"] = self.get_fullscale_current(group, line)
        if (group, line) in self._LO_IDX:
            d_ratio = self.get_divider_ratio(group, line)
            if d_ratio > 0:
                r["lo_freq"] = self.get_lo_multiplier(group, line) * 100_000_000 // d_ratio
            else:
                r["lo_freq"] = 0
        if (group, line) in self._VATT_IDX:
            r["sideband"] = self.get_sideband(group, line)
            vatt = self.get_vatt_carboncopy(group, line)
            if vatt is not None:
                r["vatt"] = vatt
        return r

    def dump_rchannel(self, group: int, rline: str, rchannel: int) -> Dict[str, Any]:
        """dumping the current configuration of the receiver channel.

        :param group: an index of a group which the channel belongs to.
        :param rline: a group-local index of a line which the target channel belongs to.
        :param rchannel: a line-local index of the target channel.
        :return: the current configuration information of the channel.
        """
        self._validate_rchannel(group, rline, rchannel)
        return {
            "fnco_freq": self.get_adc_fnco(group, rline, rchannel),
        }

    def dump_rline(self, group: int, rline: str) -> Dict[str, Any]:
        """dumping the current configuration of a receiver line.

        :param group: an index of a group which the line belongs to.
        :param rline: a group-local index of the line.
        :return: the current configuration information of the line.
        """
        self._validate_rline(group, rline)
        r: Dict[str, Any] = {}
        if (group, rline) in self._LO_IDX:
            d_ratio = self.get_divider_ratio(group, rline)
            if d_ratio > 0:
                r["lo_freq"] = self.get_lo_multiplier(group, rline) * 100_000_000 // d_ratio
            else:
                r["lo_freq"] = 0
        r["cnco_freq"] = self.get_adc_cnco(group, rline)
        r["channels"] = {
            rch: self.dump_rchannel(group, rline, rch) for rch in range(self.get_num_rchannels_of_rline(group, rline))
        }
        return r


class QubeConfigSubsystem(
    QuelMeeBoardConfigSubsystem,
):
    __slots__ = ()

    _GROUPS: Set[int] = {0, 1}

    _DAC_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (0, 1),
        (0, 2): (0, 2),
        (0, 3): (0, 3),
        (1, 0): (1, 3),
        (1, 1): (1, 2),
        (1, 2): (1, 1),
        (1, 3): (1, 0),
    }

    _LO_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (1, 0),
        (0, 2): (2, 0),
        (0, 3): (3, 0),
        (1, 0): (7, 0),
        (1, 1): (6, 0),
        (1, 2): (5, 0),
        (1, 3): (4, 0),
        (0, "r"): (0, 1),
        (0, "m"): (1, 1),
        (1, "r"): (7, 1),
        (1, "m"): (6, 1),
    }

    _MIXER_IDX: Dict[Tuple[int, int], int] = {
        (0, 0): 0,
        (0, 1): 1,
        (0, 2): 2,
        (0, 3): 3,
        (1, 0): 7,
        (1, 1): 6,
        (1, 2): 5,
        (1, 3): 4,
    }

    _VATT_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (0, 1),
        (0, 2): (0, 2),
        (0, 3): (0, 3),
        (1, 0): (0, 7),
        (1, 1): (0, 6),
        (1, 2): (0, 5),
        (1, 3): (0, 4),
    }

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)


class Quel1ConfigSubsystem(QubeConfigSubsystem, Quel1ConfigSubsystemRfswitch):
    __slots__ = ()

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        QubeConfigSubsystem.__init__(self, css_addr, boxtype, port, timeout, sender_limit_by_binding)

    def initialize(self) -> None:
        super().initialize()
        self._construct_rfswitch(0)

    def configure_peripherals(
        self,
        param: dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None:
        # Notes: init_rfswitch() should be called at the head of configure_peripherals() to avoid the leakage of sprious
        #        during the initialization of RF components.
        self.init_rfswitch(param["gpio"][0])
        super().configure_peripherals(
            param,
            ignore_access_failure_of_adrf6780=ignore_access_failure_of_adrf6780,
            ignore_lock_failure_of_lmx2594=ignore_lock_failure_of_lmx2594,
        )

    def dump_line(self, group: int, line: int) -> Dict[str, Any]:
        r = super().dump_line(group, line)
        if (group, line) in self._RFSWITCH_NAME:
            r["rfswitch"] = "block" if self.is_blocked_line(group, line) else "pass"
        return r

    def dump_rline(self, group: int, rline: str) -> Dict[str, Any]:
        r = super().dump_rline(group, rline)
        if (group, rline) in self._RFSWITCH_NAME:
            r["rfswitch"] = "loop" if self.is_blocked_line(group, rline) else "open"
        return r


class QubeOuTypeAConfigSubsystem(QubeConfigSubsystem, Quel1ConfigSubsystemNoRfswitch):
    __slots__ = ()

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {
        (0, "r"): (0, 3),
        (1, "r"): (1, 3),
    }

    # TODO: will be replaced with a parser method
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {
        (0, "r"): ((0, 5),),
        (1, "r"): ((1, 5),),
    }

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        if boxtype != Quel1BoxType.QuBE_OU_TypeA:
            raise ValueError(f"invalid boxtype: {boxtype} for {self.__class__.__name__}")
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)


class QubeOuTypeBConfigSubsystem(QubeConfigSubsystem, Quel1ConfigSubsystemNoRfswitch):
    __slots__ = ()

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {}

    # TODO: will be replaced with a parser method
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {}

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        if boxtype != Quel1BoxType.QuBE_OU_TypeB:
            raise ValueError(f"invalid boxtype: {boxtype} for {self.__class__.__name__}")
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)


class Quel1TypeAConfigSubsystem(Quel1ConfigSubsystem):
    __slots__ = ()

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {
        (0, "r"): (0, 3),
        (0, "m"): (0, 2),
        (1, "r"): (1, 3),
        (1, "m"): (1, 2),
    }

    # TODO: will be replaced with a parser method
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {
        (0, "r"): ((0, 5),),
        (0, "m"): ((0, 4),),
        (1, "r"): ((1, 5),),
        (1, "m"): ((1, 4),),
    }

    _RFSWITCH_NAME: Dict[Tuple[int, Union[int, str]], Tuple[int, str]] = {
        (0, 0): (0, "path0"),
        (0, 1): (0, "path1"),
        (0, 2): (0, "path2"),
        (0, 3): (0, "path3"),
        (0, "r"): (0, "path0"),
        (0, "m"): (0, "monitor"),
        (1, 0): (1, "path0"),
        (1, 1): (1, "path1"),
        (1, 2): (1, "path2"),
        (1, 3): (1, "path3"),
        (1, "r"): (1, "path0"),
        (1, "m"): (1, "monitor"),
    }

    _RFSWITCH_SUBORDINATE_OF: Dict[Tuple[int, Union[int, str]], Tuple[int, Union[int, str]]] = {
        (0, 0): (0, "r"),
        (1, 0): (1, "r"),
    }

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        if boxtype not in {Quel1BoxType.QuBE_RIKEN_TypeA, Quel1BoxType.QuEL1_TypeA}:
            raise ValueError(f"invalid boxtype: {boxtype} for {self.__class__.__name__}")
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)


class Quel1TypeBConfigSubsystem(Quel1ConfigSubsystem):
    __slots__ = ()

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {
        (0, "m"): (0, 2),
        (1, "m"): (1, 2),
    }

    # TODO: will be replaced with a parser method
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {
        (0, "m"): ((0, 4),),
        (1, "m"): ((1, 4),),
    }

    _RFSWITCH_NAME: Dict[Tuple[int, Union[int, str]], Tuple[int, str]] = {
        (0, 0): (0, "path0"),
        (0, 1): (0, "path1"),
        (0, 2): (0, "path2"),
        (0, 3): (0, "path3"),
        (0, "m"): (0, "monitor"),
        (1, 0): (1, "path0"),
        (1, 1): (1, "path1"),
        (1, 2): (1, "path2"),
        (1, 3): (1, "path3"),
        (1, "m"): (1, "monitor"),
    }

    _RFSWITCH_SUBORDINATE_OF: Dict[Tuple[int, Union[int, str]], Tuple[int, Union[int, str]]] = {}

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        if boxtype not in {Quel1BoxType.QuBE_RIKEN_TypeB, Quel1BoxType.QuEL1_TypeB}:
            raise ValueError(f"invalid boxtype: {boxtype} for {self.__class__.__name__}")
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)


class Quel1NecConfigSubsystem(
    QuelMeeBoardConfigSubsystem,
    Quel1ConfigSubsystemNoRfswitch,
):
    __slots__ = ()

    _GROUPS: Set[int] = {0, 1, 2, 3}

    _DAC_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (0, 2),
        (1, 0): (0, 1),
        (1, 1): (0, 3),
        (2, 0): (1, 2),
        (2, 1): (1, 0),
        (3, 0): (1, 3),
        (3, 1): (1, 1),
    }

    _LO_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (2, 0),
        (1, 0): (1, 0),
        (1, 1): (3, 0),
        (2, 0): (6, 0),
        (2, 1): (4, 0),
        (3, 0): (7, 0),
        (3, 1): (5, 0),
        (0, "r"): (0, 1),
        (1, "r"): (1, 1),
        (2, "r"): (6, 1),
        (3, "r"): (7, 1),
    }

    _MIXER_IDX: Dict[Tuple[int, int], int] = {
        (0, 0): 0,
        (0, 1): 2,
        (1, 0): 1,
        (1, 1): 3,
        (2, 0): 6,
        (2, 1): 4,
        (3, 0): 7,
        (3, 1): 5,
    }

    _VATT_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (0, 2),
        (1, 0): (0, 1),
        (1, 1): (0, 3),
        (2, 0): (0, 6),
        (2, 1): (0, 4),
        (3, 0): (0, 7),
        (3, 1): (0, 5),
    }

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {
        (0, "r"): (0, 3),
        (1, "r"): (0, 2),
        (2, "r"): (1, 3),
        (3, "r"): (1, 2),
    }

    # TODO: will be replaced with a parser method
    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {
        (0, "r"): ((0, 5),),
        (1, "r"): ((0, 4),),
        (2, "r"): ((1, 5),),
        (3, "r"): ((1, 4),),
    }

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeSockClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeSockClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        if boxtype != Quel1BoxType.QuEL1_NEC:
            raise ValueError(f"invalid boxtype: {boxtype} for {self.__class__.__name__}")
        super().__init__(css_addr, boxtype, port, timeout, sender_limit_by_binding)



from ._config import Box, DeskewConfiguration, Port, load_default_configuration
from ._e7awg_delay_compensator import (
    E7awgDelayCompensator,
    WaitAmountResolver,
    extract_wave_dict,
    extract_wave_list,
    register_blank_wavedata,
)
from ._stable_count_proposer import StableCountProposer

__all__ = [
    "Box",
    "DeskewConfiguration",
    "E7awgDelayCompensator",
    "Port",
    "StableCountProposer",
    "WaitAmountResolver",
    "extract_wave_dict",
    "extract_wave_list",
    "load_default_configuration",
    "register_blank_wavedata",
]
